from pandas import DataFrame

from .. import _error_messages
from .._data_objects import DataPoints, Investments
from .._exceptions import ResourceNotFoundError


def _get_investment_ids(investment_object: Investments) -> list:
    investment_id_list = investment_object.get_investment_ids()
    if not investment_id_list:
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_INVESTMENT_DATA)
    return investment_id_list


def _get_data_points(investment_object: Investments, data_point_object: DataPoints, display_name: bool = False) -> DataFrame:
    data_point_object.validate_data_point_ids()
    return data_point_object.get_data_points(
        list_id=investment_object.list_id, search_criteria_id=investment_object.search_criteria, display_name=display_name
    )


def _get_data_point_col_names(data_points: DataFrame, data_point_alias_to_cols: dict) -> list:
    col_names = []
    for _, data_point in data_points.iterrows():
        alias = data_point.get("alias", "")
        columns = data_point_alias_to_cols.get(alias, [])
        col_names.extend(columns)
    return col_names


# def _build_result_data_frame(investment_data_list: list) -> DataFrame:
#     result = DataFrame(investment_data_list)
#     # TODO : .where() with Pandas 1.1.5 has a bug which converts all the columns including float to object
#     # this was solved in Pandas update 1.3.5. Until we update Pandas, we will have to use this _replace_null_values()
#     # as a workaround
#     # results = df.where(result.notnull(), None)
#     return result



import logging
import os
import pathlib
import uuid
from typing import Dict

import boto3
import jwt
import requests
import simplejson as json
from urllib3.exceptions import InsecureRequestWarning

from ..direct._backend_apis import DeliveryAPIBackend
from ..direct._config import _Config
from ..direct._exceptions import ForbiddenError, UnauthorizedDeliveryException

# NOTE: We are not able to verify the SSL cert on MDS API since
#       the VPC endpoint domain name is not on the server's SSL cert.
#       So with a heavy heart, we disable the InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


_logger = logging.getLogger(__name__)
_config = _Config()


def _get_user_id_from_uim_token() -> str:
    """Returns the user's id by decoding their UIM token"""
    try:
        md_auth_token = md_auth_token = os.getenv("MD_AUTH_TOKEN", os.getenv("UIM_JWT", ""))
        decode_key = os.getenv("UIM_DECODE_KEY", "https://login-prod.morningstar.com/.well-known/jwks.json")
        jwks_client = jwt.jwks_client.PyJWKClient(decode_key)
        signing_key = jwks_client.get_signing_key_from_jwt(md_auth_token)
        decoded_jwt = jwt.decode(
            md_auth_token,
            signing_key.key,
            algorithms=["RS256"],
            options={"verify_signature": True, "verify_exp": True, "verify_aud": False},
            leeway=60,
        )
        user_id: str = decoded_jwt["https://morningstar.com/mstar_id"]
        return user_id.upper()
    except Exception as err:
        _logger.error(f"Error decoding MD_AUTH_TOKEN: {err}")
        raise


def _get_user_id_from_eams_token() -> str:
    """Returns the user's id by decoding their EAMS token"""
    try:
        key = os.getenv("EAMS_PUBLIC_KEY", "").replace("\\n", "\n")
        token = os.getenv("EAMS_JWT")
        decoded = jwt.decode(token, key, algorithms=["RS256"])
        user_id: str = decoded["sub"]
        return user_id.upper()
    except Exception as err:
        _logger.error(f"Error decoding EAMS token: {err}")
        raise


def _get_user_id() -> str:
    """Returns the user's id using their UIM token if it exists, EAMS token otherwise"""
    try:
        user_id = _get_user_id_from_uim_token()
        _logger.info(f"Using UIM token for User ID: {user_id}")
        return user_id
    except Exception:
        pass

    user_id = _get_user_id_from_eams_token()
    _logger.info(f"Using EAMS token for User ID: {user_id}")
    return user_id


def _get_job_name() -> str:
    """Returns the job name for a feed or delivery"""
    return str(uuid.uuid4())


def _get_request_id() -> str:
    """Returns the request id for use in POST requests"""
    request_id = os.getenv("REQUEST_ID", str(uuid.uuid4()))
    _logger.info(f"RequestId: {request_id} passed to POST requests")
    return request_id


def _get_bucket_name() -> str:
    """Returns the mds bucket name for s3 upload"""
    return os.getenv("MDS_BUCKET_NAME", "velo-notebook-output-prod-us-east-1")


def _create_feed_body() -> Dict:
    """Creates the post body which is used to call the MDS api when creating a feed id"""
    body = {
        "userId": _get_user_id(),
        "userName": _get_user_id(),
        "jobName": _get_job_name(),
        "jobMetaData": {
            "jobRunUrl": "place_holder",
            "jobRunMethod": "POST",
            "jobPostBody": {
                "app_source": "Morningstar Data Service(MDS)",
                "run_type": "notebook",
                "notebook_parameters": {"bucket": _get_bucket_name()},
            },
        },
    }
    return body


def _get_mds_base_url() -> str:
    return f"{_config.mds_api_url()}"


def _get_al_proxy_base_url() -> str:
    return f"{_config.al_proxy_api_url()}"


def _encrypt_ftp_password(password: str) -> str:
    """Takes a password and encrypts it using the user_id as the salt"""
    _api_backend = DeliveryAPIBackend()
    url = f"{_get_al_proxy_base_url()}encrypt-message"
    data = {"message": password, "salt": _get_user_id()}
    try:
        _logger.info(f"Calling {url} to encrypt ftp password")
        response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
        encrypted_password: str = response_json["encrypted_message"]
        return encrypted_password
    except Exception:
        _logger.error("Could not encrypt password using /encrypt-message")
        raise


def _create_output_file_name(file_path: str) -> str:
    """Creates the output file name for the file which will be uploaded to MDS s3 bucket"""
    file_name = pathlib.Path(file_path).name
    mds_request_id = os.getenv("FEED_RUN_ID", str(uuid.uuid4()))
    return f"{mds_request_id}-{file_name}"


def _retrieve_credentials() -> Dict[str, str]:
    """Retrieves the credentials required to access the MDS s3 bucket"""
    _api_backend = DeliveryAPIBackend()
    url = f"{_get_al_proxy_base_url()}assume-mds-role?user_id={_get_user_id()}"
    try:
        _logger.info(f"Calling {url} to get assume role credentials")
        response_json: Dict[str, str] = _api_backend.do_get_request(url)
        return response_json
    except ForbiddenError:
        _logger.error(f"Unable to retrieve credentials for MDS s3 bucket upload")
        _logger.error(f"User does not have access to deliver function feature")
        raise UnauthorizedDeliveryException from None


def _get_mds_bucket() -> boto3.session.Session.resource:
    """Returns the MDS bucket where we will upload the file before calling the delivery endpoint"""
    credentials = _retrieve_credentials()
    s3_resource = boto3.resource(
        "s3",
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )
    bucket = s3_resource.Bucket(_get_bucket_name())
    return bucket



import logging
import os
import uuid
from typing import Any, Dict, Optional, Union

import requests

from ...direct._base_api import APIBackend
from ...direct._exceptions import (
    BadRequestException,
    NetworkExceptionError,
    UnauthorizedDeliveryException,
)
from .._core_api import make_request

_logger = logging.getLogger(__name__)


class DeliveryAPIBackend(APIBackend):
    def do_post_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a POST request
        """
        return self._make_request(url, "POST", data)

    def _get_request_id(self) -> str:
        """Returns the request id for use in POST requests"""
        request_id = os.getenv("FEED_RUN_ID", str(uuid.uuid4()))
        _logger.info(f"RequestId: {request_id} passed to POST requests")
        return request_id

    def _make_request(self, url: str, method: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        _logger.info(f"Requesting: {method} {url}")
        try:
            headers = {**self.get_headers(), "X-API-RequestId": self._get_request_id()}
            res = make_request(method, url, headers=headers, verify=False, data=data)
            res.raise_for_status()
            return res.json()
        except requests.ConnectionError:
            raise NetworkExceptionError from None
        except requests.HTTPError:
            self._handle_custom_http_errors(res)
            self._handle_default_http_errors(res)

    def _handle_custom_http_errors(self, res: Any) -> Any:
        try:
            response_message = res.json().get("status")
        except requests.JSONDecodeError:
            response_message = res.text

        if res.status_code == 401:
            _logger.debug(f"Unauthorized Error: {res.status_code} {response_message}")
            raise UnauthorizedDeliveryException from None
        if res.status_code == 400:
            _logger.debug(f"Bad Request: {res.status_code} {response_message}")
            raise BadRequestException(message=response_message) from None



from typing import Any, Optional

import pandas as pd
from pandas import DataFrame

from ..._base import _logger
from .. import _decorator, _error_messages
from .._base_api import APIBackend
from .._config import _Config
from .._error_messages import (
    BAD_REQUEST_INVALID_PORTFOLIO_NAME,
    BAD_REQUEST_INVALID_PORTFOLIO_TYPE,
    BAD_REQUEST_PORTFOLIO_NAME_ALREADY_EXISTS,
    CLIENT_ACCOUNTS_PORTFOLIO_TYPE_USED,
    RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_List,
)
from .._exceptions import (
    AccessDeniedError,
    BadRequestException,
    InternalServerError,
    ResourceNotFoundError,
)
from .._portfolio import HoldingsEntityRequestBuilder, PortfolioSetting, PortfolioType
from .._portfolio._common import PortfolioDataApiBackend, PortfolioNotFoundError

_config = _Config()
_portfolio_api = PortfolioDataApiBackend()


@_decorator.not_null
@_decorator.typechecked
def get_portfolios(portfolio_type: Optional[str] = None) -> DataFrame:
    """Returns all portfolios for the given portfolio type.

    Args:
        portfolio_type(:obj:`str`): The portfolio type. The available options are: model_portfolios, custom_benchmarks.

    :Returns:
        DataFrame: A DataFrame object with portfolio details. DataFrame columns include:

        * PortfolioId
        * Name
        * Type

    :Examples:
        Get client account portfolios.
    ::

        import morningstar_data as md

        df = md.direct.user_items.get_portfolios(portfolio_type = "model_portfolios")
        df

    :Output:
        =======================================     ==========    ===============
        PortfolioId                                 Name          Type
        =======================================     ==========    ===============
        3d25b613-ae42-4cb6-bd05-7ba7c74883f3;MD     Portfolio1    model_portfolios
        5d25b613-ae42-4cb6-bd05-7ba7c74883f3;MD     Portfolio2    model_portfolios
        ...
        =======================================     ==========    ===============

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    all_data_frame = DataFrame()
    if portfolio_type == "client_accounts":
        raise BadRequestException(CLIENT_ACCOUNTS_PORTFOLIO_TYPE_USED)
    if portfolio_type is None:
        types = ",".join([PortfolioType.model_portfolios.key, PortfolioType.custom_benchmarks.key])
    else:
        types = _portfolio_type_map(portfolio_type)
    try:
        url = f"{_config.portfolio_service_url()}/portfoliodataservice/v3/portfolios?types={types}"
        _logger.debug(f"Fetching portfolio data from: {url}")
        response_json = _portfolio_api.do_get_request(url)
        df = DataFrame(response_json["snapshots"])
        new_data_frame = df.drop(columns=["dataPoints", "permission", "folderId"])
        new_data_frame.columns = ["PortfolioId", "Name", "Type"]
        new_data_frame["Type"] = new_data_frame["Type"].map(_mapping_portfolio_type_full_name)
        if portfolio_type is not None:
            _logger.debug(f"Portfolio type is {portfolio_type}. Filtering data to match this portfolio.")
            new_data_frame = new_data_frame.loc[new_data_frame["Type"] == portfolio_type]
        all_data_frame = pd.concat([all_data_frame, new_data_frame])
    except PortfolioNotFoundError:
        _logger.debug(f"There are no portfolio(type={types}) in the Direct account.")
    except AccessDeniedError as e:
        raise AccessDeniedError(str(e)) from None
    except BadRequestException as e:
        raise BadRequestException(str(e)) from None
    except Exception as e:
        raise InternalServerError(str(e)) from None

    if all_data_frame.empty:
        if len(types) == 2:  # i.e. if only one portfolio type was selected
            _logger.debug(f"Requested resource not found for portfolio type `{portfolio_type}`")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ACCOUNT) from None
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ACCOUNTS) from None
    else:
        all_data_frame = all_data_frame.reset_index(drop=True)

    return all_data_frame


def _portfolio_type_map(portfolio_type: str) -> str:
    try:
        return PortfolioType[portfolio_type].key
    except KeyError:
        raise BadRequestException(BAD_REQUEST_INVALID_PORTFOLIO_TYPE)


def _mapping_portfolio_type_full_name(val: str) -> str:
    return PortfolioType.get_full_name_by_abbr(val)


@_decorator.typechecked
def save_portfolio(
    portfolio_name: str, portfolio_type: PortfolioType, holdings: Optional[DataFrame] = None, overwrite_if_exists: bool = False
) -> DataFrame:
    """Save or update a portfolio.

    Args:
        portfolio_name (:obj:`str`): Name of the portfolio.
        portfolio_type (:obj:`PortfolioType`):Type of the portfolio. Supported portfolio types: model_portfolios, custom_benchmarks.
        overwrite_if_exists(:obj:`bool`): If true, an existing portfolio will be overwritten with new name and holdings.
        holdings:(:obj:`DataFrame`, `optional`) : A DataFrame object containing holdings to save to this portfolio.

    :Returns:
        DataFrame: A DataFrame object, columns include:

        * Name
        * Portfolio Id

    :Examples:

    Save a new portfolio.
    ::

        import pandas as pd
        import morningstar_data as md
        from morningstar_data.direct.user_items.portfolio import PortfolioType

        holdings_data_frame = pd.DataFrame(
            [
                ["FOUSA00DFS;FO", "2020-10-31", 33.33],
                ["0P000002RH;ST", "2020-10-31", 66.67]
            ],
            columns=["HoldingId", "Portfolio Date", "Weight"],
        )

        df = md.direct.user_items.save_portfolio(
            portfolio_name = "new_portfolio",
            portfolio_type = PortfolioType.model_portfolios,
            holdings = holdings_data_frame,
            overwrite_if_exists = False,
        )

        df

    :Output:
        ===============  ====================================
        Name             Portfolio Id
        ===============  ====================================
        new_portfolio    295F7E59-15AC-4424-958E-3BD8B0A733EE
        ===============  ====================================

        Update an existing portfolio.

        ::

            import pandas as pd
            import morningstar_data as md
            from morningstar_data.direct.user_items.portfolio import PortfolioType

            # Updated holdings data
            holdings_data_frame = pd.DataFrame(
                [
                ["FOUSA00DFS;FO", "2020-05-31", 50],
                ["0P000002RH;ST", "2020-05-31", 50]
                ],
                columns=["HoldingId", "Portfolio Date", "Weight"]
            )

            df = md.direct.user_items.save_portfolio(
                portfolio_name = "new_portfolio",
                portfolio_type = PortfolioType.model_portfolios,
                holdings = holdings_data_frame,
                overwrite_if_exists = True # Update the existing portfolio
            )

            df


        =============== ====================================
        Name             Portfolio Id
        =============== ====================================
        new_portfolio   295F7E59-15AC-4424-958E-3BD8B0A733EE
        =============== ====================================

    """
    _logger.info("Validating portfolio name")
    if portfolio_name is None or portfolio_name.strip() == "":
        _logger.debug(f"Invalid portfolio name: {portfolio_name}. Raising exception.")
        raise BadRequestException(BAD_REQUEST_INVALID_PORTFOLIO_NAME) from None

    _logger.info("Validating portfolio type")
    if portfolio_type not in [PortfolioType.custom_benchmarks, PortfolioType.model_portfolios]:
        _logger.debug(f"Invalid portfolio type: {portfolio_type}")
        raise BadRequestException(BAD_REQUEST_INVALID_PORTFOLIO_TYPE) from None

    _logger.info("Validating portfolio holdings")
    if holdings is not None:
        _logger.info("Holdings is not None, creating holdings entity request builder ,and validate holdings.")
        holdings_entitys_builder = HoldingsEntityRequestBuilder(holdings)

    _logger.info("Validating against existing portfolio names.")

    _logger.info(f"Get all user created portfolios")
    user_portfolios = get_portfolios(portfolio_type.name)

    portfolio_id = None
    if (portfolio_name in user_portfolios.values) and overwrite_if_exists == False:
        # Portfolio with same name exists and user is not overwriting it. This is an invalid case.

        _logger.info("Portfolio name already exists and overwrite is set to False. Throwing error.")
        _logger.debug(f"Portfolio name: {portfolio_name}. Overwrite if exists: {overwrite_if_exists}")
        raise BadRequestException(BAD_REQUEST_PORTFOLIO_NAME_ALREADY_EXISTS) from None

    if (portfolio_name in user_portfolios.values) and overwrite_if_exists:
        # Portfolio with same name exists and user is overwriting it. This is an valid case.

        _logger.info("Portfolio name already exists and overwrite is set to True")
        _logger.info("Fetching matching portfolio id.")
        portfolio_id = user_portfolios.loc[user_portfolios["Name"].str.lower() == portfolio_name.lower(), "PortfolioId"].iloc[0]

        _logger.debug(f"Got matching portfolio id: {portfolio_id}")

    if portfolio_id is None:
        _logger.info("Portfolio id is None, so creating a new portfolio")
        # create new portfolio with default settings.
        portfolio_setting = PortfolioSetting(name=portfolio_name, portfolio_type=portfolio_type.key, portfolio_id=portfolio_id)

        portfolio_setting.save_portfolio_settings()  # Fix this in DO API to return portfolio id

        # The above statement does not return newly created portfolio id.
        # So we will call get_portfolios one more time to get the newly created portfolio id.
        _logger.info("Fetching user portfolio that match the newly created portfolio name.")
        user_portfolios = get_portfolios(portfolio_type.name)

        if portfolio_name in user_portfolios.values:
            portfolio_id = user_portfolios.loc[user_portfolios["Name"] == portfolio_name, "PortfolioId"].iloc[0]
            _logger.debug(f"Got newly created portfolio id: {portfolio_id}")

    # Assign portfolio id to the holdings and update portfolio
    if holdings_entitys_builder is not None:
        _logger.info("Holdings entity builder is not None, updating portfolio with holdings.")
        holdings_entitys_builder.portfolio_id = portfolio_id
        holdings_entitys_builder.update_portfolio_holdings()

    data = {"Name": [portfolio_name], "Portfolio Id": [portfolio_id]}

    return DataFrame(data)



import datetime
import inspect
import logging
import os
import uuid
from contextvars import Token
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import requests
import simplejson as json
from pandas import DataFrame

from .direct._api import _direct_api_request
from .direct._config import _Config

_config = _Config()
_logger = logging.getLogger(__name__)


def get_parameter(key: str) -> Optional[Any]:
    """Fetch a parameter from query string with a fallback to an environment variable

    Voila supports query string parameters (available via QUERY_STRING environment variable).
    This function helps to prioritize query string parameters over environment variables with matching names.
    """

    try:
        from os import environ as _env
        from urllib.parse import parse_qs

        # Voila supports
        query_string = _env.get("QUERY_STRING", "")
        parameters: Dict[str, Any] = parse_qs(query_string)
        if key in parameters:
            return parameters[key][0]
        if key in _env:
            return _env.get(key)
        return None
    except BaseException:
        _logger.warn(f"Something is wrong with key={key}")
        return None


def load_remote(target_folder: str, remote_endpoint: str) -> Any:
    from os import path as _path
    from subprocess import run

    try:
        result = run(
            [
                "sh",
                _path.join(_path.dirname(__file__), "remote.sh"),
                target_folder,
                remote_endpoint,
            ],
            shell=False,
            capture_output=True,
        )
        return result
    except Exception as _ex:
        _logger.warn(_ex)


def format_analytics_logs(
    function: str, params: str, component: str = "morningstar_data.direct", action: str = "FUNCTION_RUN"
) -> str:
    request_id = str(uuid.uuid1())
    result = {
        "object_type": "MD Package",
        "object_id": function,
        "application": "Analytics Lab",
        "component": component,
        "action": action,
        "session_id": os.getenv("ANALYTICSLAB_SESSION_ID"),
        "user_id": os.getenv("UIM_USER_ID"),
        "details": params,
        "event_id": request_id,
    }

    return f"{json.dumps(result, ignore_nan=True)}"


def get_log_flag(func: Callable) -> bool:
    # log_flag allows to know if a function should be logged for analytics or not
    log_flag = True

    # Return a list of frame records for the caller function’s stack. The first entry in the returned list represents the caller; the last entry represents the outermost call on the stack.
    stack = inspect.stack()

    for stk in stack:
        # <module> represents it is for function called by function execution) - should return true
        if stk.function == "<module>":
            return log_flag
        # Checks if caller is any internal function or not get_log_flag (self), sets flag to false
        elif stk.function not in ["get_log_flag", "wrapper"] or stk.function.startswith("_"):
            log_flag = False
            return log_flag
        else:
            continue

    return log_flag


def _get_user_cells_quota() -> Dict:
    url = f"{_config.asset_service_url()}/auto/analyticslab/log"
    response_json = _direct_api_request("get", url)
    user_cells_quota = {"daily_cell_limit": response_json["max"], "daily_cell_remaining": response_json["remains"]}
    return user_cells_quota


def _get_data_points_total_columns(data_point_details: list) -> int:
    url = f"{_config.data_point_service_url()}v1/datapoints/columns"
    columns_response_json: List[Any] = _direct_api_request("POST", url, json.dumps(data_point_details, ignore_nan=True))
    total_columns = sum(item.get("columns") for item in columns_response_json)
    return total_columns


def join_url(*parts: str) -> str:
    return "/".join(part.strip("/") for part in parts)


def data_point_dataframe_to_list(dp_dataframe: DataFrame) -> List[Dict[str, Any]]:
    dp_dict = dp_dataframe.replace(np.nan, None).to_dict(orient="records")

    # Avoid sending dp settings with None values, the API will consider them invalid
    return [{k: v for k, v in dp.items() if v is not None} for dp in dp_dict]



from .._portfolio import PortfolioType
from .data_set import get_data_set, get_data_set_details, get_data_sets
from .investment_lists import (
    get_investment_list,
    get_investment_lists,
    save_investment_list,
)
from .portfolio import get_portfolios, save_portfolio
from .search_criteria import (
    get_search_criteria,
    get_search_criteria_condition,
    get_search_criteria_conditions,
    get_search_criterias,
    get_search_results,
)



from enum import Enum
from typing import Any, Dict, Optional

from pandas import DataFrame, concat

from ._base import _logger
from .direct import _decorator
from .direct._base_api import APIBackend
from .direct._config import _Config
from .direct._error_messages import (
    RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD,
    RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD_UNIVERSE,
)
from .direct._exceptions import ResourceNotFoundError

_config = _Config()


class _LookupParameter(Enum):
    KEYWORD = 0
    KEYWORD_UNIVERSE = 1


class LookupAPIBackend(APIBackend):
    """
    Subclass to call the Holding API and handle any HTTP errors that occur.
    """

    def __init__(self, lookup_parameter: _LookupParameter) -> None:
        self._lookup_parameter = lookup_parameter
        super().__init__()

    def _get_resource_not_found_message(self) -> str:
        if self._lookup_parameter == _LookupParameter.KEYWORD:
            return RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD
        return RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD_UNIVERSE

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(self._get_resource_not_found_message()) from None


_keyword_api_request = LookupAPIBackend(_LookupParameter.KEYWORD)
_keyword_universe_api_request = LookupAPIBackend(_LookupParameter.KEYWORD_UNIVERSE)


def _get_data_point_options(data_point: str, universe: Optional[str]) -> dict:
    url = f"{_config.data_point_service_url()}v1/datapoints/{data_point}/searchoperatorsandoptions"
    _lookup_api_request = _keyword_api_request
    if universe:
        _logger.info("Universe is specified, so searching within this universe only.")
        # Some data points need to specify universe, for example OF003(Morningstar Category).
        url = url + f"?universe={universe}"
        _lookup_api_request = _keyword_universe_api_request
    response_json: Dict[Any, Any] = _lookup_api_request.do_get_request(url)
    return response_json


def _search_option_data(data_point: str, keyword: Optional[str], universe: Optional[str] = None) -> DataFrame:
    # get the available values for special data point.
    data_point_options = _get_data_point_options(data_point, universe)
    df = DataFrame({"id": [], "name": []})
    if data_point_options:
        result = DataFrame(data=data_point_options[0].get("options"))
        result.rename(columns={"value": "id"}, inplace=True)
        df = result[["id", "name"]]

    return _filter_data_frame(df, keyword) if keyword else df


def _filter_data_frame(df: DataFrame, keyword: str) -> DataFrame:
    filtered_df = concat([df[df["id"].str.contains(keyword, case=False)], df[df["name"].str.contains(keyword, case=False)]])
    return filtered_df.drop_duplicates(subset=["id", "name"], keep="first")


def currency_codes(keyword: Optional[str] = None) -> DataFrame:
    """Returns currency codes and currency name that match the given keyword. If no keyword is provided, the function returns all currency codes and currency names.

    Args:
        keyword (Optional[str], optional): A string used to lookup currency codes. Example: "USD". Returns matching currency code for the keyword 'USD'.

    :Returns:
        DataFrame: Returns a DataFrame. The DataFrame columns include:

        * currency_code
        * currency_name

    :Examples:

        Search currency codes based on keyword "Afgh"
    ::

        import morningstar_data as md
        df = md.lookup.currency_codes(
            keyword="Afgh"
        )
        df

    :Output:
        ==============  ===============
        currency_code      currency_name
        ==============  ===============
        AFN                Afghani
        ==============  ===============

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when a currency code matching the given keyword does not exist.
    """

    # LS05M is a datapoint which name is currency, and its options include all available currency value.
    currency_df = _search_option_data(data_point="LS05M", keyword=keyword)
    currency_df.rename(columns={"id": "currency_code", "name": "currency_name"}, inplace=True)
    return currency_df



from typing import Any, Dict, List, Optional, Union

import requests

# NOTE! We are not able to verify the SSL cert on Direct Search API since
#       the VPC endpoint domain name is not on the server's SSL cert.
#       So with a heavy heart, we disable the InsecureRequestWarning
from urllib3.exceptions import InsecureRequestWarning

from .._base import _logger
from ._config import _Config
from ._core_api import make_request
from ._exceptions import ApiRequestException, ApiResponseException, QueryLimitException

_config = _Config()

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def _direct_api_request(method: str, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
    """
    Generic request method for the Direct API which covers headers and aspects
    of response handling that are consistent across API calls

    :method: Request method, eg. GET or POST
    :url: Request URL
    :data: (optional) POST message body, if any
    """
    headers = _config.get_headers()
    _logger.debug(f"Requesting: {method} {url} with headers: {headers}")
    try:
        res = make_request(method, url, headers=headers, verify=False, data=data)
        res.raise_for_status()

        return res.json()
    except requests.ConnectionError as e:
        _logger.error(repr(e))
        raise ApiRequestException(f"Failed connecting to {url}")
    except requests.HTTPError as e:
        message = f"Received {res.status_code} error from {url}"
        json_response = res.json()
        if res.status_code == 403 and "Exceed query limitation." in json_response["message"]:
            query_limit = _get_security_data_query_limit()
            raise QueryLimitException(query_limit) from None
        else:
            if "_meta" in json_response and "hint" in json_response["_meta"]:
                message += f" ({json_response['_meta']['hint']})"
            elif "message" in json_response:
                message += f" ({json_response['message']})"
            raise ApiResponseException(message, res.status_code)


def _get_security_data_query_limit() -> str:
    try:
        headers = _config.get_headers()
        url = f"{_config.securitydata_service_url()}v1/limitation/summary"

        res = make_request("GET", url, headers=headers, verify=False)
        json_response = res.json()
        return str(json_response["limitationTotal"])
    except:
        return ""



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from .. import mdapi
from .._base import _logger
from .._utils import (
    _get_data_points_total_columns,
    _get_user_cells_quota,
    data_point_dataframe_to_list,
)
from ..mdapi import RequestObject
from . import _decorator, _error_messages, _utils
from ._base_api import APIBackend
from ._config import _Config
from ._data_objects import Investments
from ._data_point import _request_asset_flow_data_points
from ._data_type import DryRunResults
from ._exceptions import BadRequestException, QueryLimitException, ResourceNotFoundError

_config = _Config()


class AssetFlowAPIBackend(APIBackend):
    """
    Subclass to call the Asset Flow Data API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 403 and "Exceed query limitation." in response_message:
            _logger.debug(f"Query Limit Exception: {res.status_code} {response_message}")
            query_limit = self._get_security_data_query_limit()
            raise QueryLimitException(query_limit) from None
        elif res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ASSET_FLOW) from None

    def _get_security_data_query_limit(self) -> str:
        try:
            url = f"{_config.securitydata_service_url()}v1/limitation/summary"
            res = self.do_get_request(url)
            return str(res["limitationTotal"])
        except Exception as e:
            _logger.error(f"Error getting security data query limit: {e}")
            raise QueryLimitException from None


_asset_flow_api_request = AssetFlowAPIBackend()


@_decorator.typechecked
def get_asset_flow_markets() -> DataFrame:
    """

    Returns:

        Returns all investment markets that can be used to retrieve asset flow data. For example, "US Open-end & ETFs ex MM ex FoF".

        DataFrame: A DataFrame object with asset flow markets data. DataFrame columns include:

        * marketId
        * marketName
        * currency

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.get_asset_flow_markets()
        df

    :Output:
        ========  ===============================  ========
        marketId  marketName                       currency
        ========  ===============================  ========
        5         US Open-end & ETF ex MM ex FoF   USD
        6         US Open-end, ETF, and MM ex FoF  USD
        ...
        ========  ===============================  ========

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    url = f"{_config.securitydata_service_url()}v1/assetflow/all-markets"
    response_json = _asset_flow_api_request.do_get_request(url)
    if response_json and isinstance(response_json, list):
        return DataFrame(response_json)[["marketId", "marketName", "currency"]]
    else:
        return DataFrame({"marketId": [], "marketName": [], "currency": []})


def get_asset_flow_data_points() -> DataFrame:
    """

    Returns:

        Returns all available data points related to asset flows.

        DataFrame: A DataFrame object with asset flow data points data. DataFrame columns include:

        * datapointId
        * datapointName
        * asOfDate
        * alias
        * startDate
        * endDate
        * frequency

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.get_asset_flow_data_points()
        df

    :Output:
        ===========  =============  ==========  ==========================================  =========  =======  =========
        datapointId  datapointName  asOfDate    alias                                       startDate  endDate  frequency
        ===========  =============  ==========  ==========================================  =========  =======  =========
        TNA0M        XXX            2021-09-30  Total Net Assets-Market Value(Share Class)  None       None     None
        ...
        ===========  =============  ==========  ==========================================  =========  =======  =========

    """
    response_json = _request_asset_flow_data_points()
    if response_json and isinstance(response_json, list):
        for settings in response_json:
            settings["datapointName"] = settings.pop("name")
            settings["alias"] = settings["datapointName"]

        settings = _utils._extract_data(response_json)
        settings_data_frame = DataFrame(settings)

        column_list = settings_data_frame.columns.tolist()
        if ("datapointId" in column_list) & (column_list.index("datapointId") != 0):
            data_point_id_col = settings_data_frame["datapointId"]
            settings_data_frame = settings_data_frame.drop("datapointId", axis=1)
            settings_data_frame.insert(0, "datapointId", data_point_id_col)

        column_list = settings_data_frame.columns.tolist()
        if ("datapointName" in column_list) & (column_list.index("datapointName") != 1):
            data_point_name_col = settings_data_frame["datapointName"]
            settings_data_frame = settings_data_frame.drop("datapointName", axis=1)
            settings_data_frame.insert(1, "datapointName", data_point_name_col)

        settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)
        return settings_data_frame
    else:
        return DataFrame()


InvestmentsParam = Union[List[str], str, Dict[str, Any]]


@dataclass
class AssetFlowRequest(RequestObject):
    market_id: str
    datapoints: List[Dict[str, Any]]
    investments: Optional[InvestmentsParam]


@_decorator.typechecked
def get_asset_flow(
    market_id: str,
    data_point_settings: DataFrame,
    investments: Optional[Union[List[str], str, Dict[str, Any]]] = None,
    dry_run: Optional[bool] = False,
) -> Union[DataFrame, DryRunResults]:
    """Get asset flow data for a market of investments or specific investments within a market

    Args:
        market_id(:obj:`str`): A numeric code representing a broad market of investments. For example, the code for "US Open-end & ETF ex MM ex FoF" is "5". Use the `get_asset_flow_markets <./assetflow.html#morningstar_data.direct.get_asset_flow_markets>`_ function to retrieve a full list of codes.

        data_point_settings(:obj:`DataFrame`): A DataFrame of data points with defined settings. Each row represents a data point. Each column is a configurable setting. This DataFrame can be obtained by `retrieving asset flow data points <./assetflow.html#morningstar_data.direct.get_asset_flow_data_points>`_, or by `retrieving data point settings <./lookup.html#morningstar_data.direct.get_data_point_settings>`_.

        investments(:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section of `get_investment_data <./investment.html#morningstar_data.direct.get_investment_data>`_ or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        dry_run(:obj:`bool`, `optional`): When True, the query will not be executed. Instead, a DryRunResults object will be returned with details about the query's impact on daily cell limit usage. When True, the 'investments' parameter must be a list containing at least one element.

    :Returns:

        There are two return types:

        * DataFrame: A DataFrame object with asset flow data. DataFrame columns include ``investmentId`` and data point names, as
        provided in ``data_point_settings``.

        * DryRunResults: Returned if dry_run=True is passed

          * estimated_cells_used: Number of cells by this query
          * daily_cells_remaining_before: How many cells are remaining in your daily cell limit before running this query
          * daily_cells_remaining_after: How many cells would be remaining in your daily cell limit after running this query
          * daily_cell_limit: Your total daily cell limit

    :Examples:

    ::

        import morningstar_data as md
        import pandas

        ASSET_FLOW_DATA_POINT_SETTINGS = [
            {
                "datapointId": "TNA0M",
                "datapointName": "Total Net Assets-Market Value(Share Class)",
                "asOfDate": "2021-08-30",
                "alias": "Total Net Assets-Market Value(Share Class)",
                "startDate": None,
                "endDate": None,
                "frequency": None,
            }
        ]
        settings = pandas.DataFrame(ASSET_FLOW_DATA_POINT_SETTINGS)
        df = md.direct.get_asset_flow(
            investments=["F000010HRO"], market_id="165", data_point_settings=settings
        )
        df

    :Output:
        ============  =======================================================
        investmentId  Total Net Assets-Market Value(Share Class) - 2021-06-30
        ============  =======================================================
        F000010HRO    0.00188
        ============  =======================================================

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if investments == {}:
        raise mdapi.BadRequestError("investments may not be an empty search criteria")

    if data_point_settings.empty:
        raise mdapi.BadRequestError("data_point_settings must not be empty")

    if dry_run and not investments:
        raise BadRequestException(_error_messages.BAD_REQUEST_ERROR_NO_INVESTMENT_IDS_FOR_DRY_RUN)

    if dry_run:
        if investments is not None:
            investment_object = Investments(investments)
            investment_id_list = investment_object.get_investment_ids()

        data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
        data_point_list = data_point_settings.to_dict(orient="records")
        estimated_cells_used = _get_data_points_total_columns(data_point_list) * len(investment_id_list)
        user_cells_quota = _get_user_cells_quota()
        dry_run_results = DryRunResults(
            estimated_cells_used=estimated_cells_used,
            daily_cells_remaining_before=user_cells_quota["daily_cell_remaining"],
            daily_cell_limit=user_cells_quota["daily_cell_limit"],
        )
        return dry_run_results

    datapoints = data_point_dataframe_to_list(data_point_settings)
    # TODO: Any extra validation? Check investment list not too long? Or do we let exception happen and catch it
    return mdapi.call_remote_function(
        "get_asset_flow",  # TODO: Make this a constant/enum
        AssetFlowRequest(
            market_id=market_id,
            datapoints=datapoints,
            investments=investments,
        ),
    )



ALL_ASSET_FLOW_DATA_POINTS = [
    "TNA0M",
    "TNA0R",
    "ENF0M",
    "ENF0R",
    "OGR0R",
    "MAP0R",
    "TNA1M",
    "TNA1R",
    "ENF1M",
    "ENF1R",
    "OGR1R",
    "MAP1R",
    "ENFM1",
    "ENFQD",
    "ENFYD",
    "ENFTM",
    "ENF24",
    "ENF36",
]

FORMAT_DATE = "%Y-%m-%d"

# Some function require these data point settings in order to successfully execute without errors.
REQUIRED_DATA_POINT_SETTINGS = [
    "datapointId",
    "datapointName",
    "isTsdp",
    "alias",
    "displayName",
]

VISIBLE_DATA_POINT_SETTINGS = [
    "currency",
    "preEuroConversion",
    "sourceId",
    "frequency",
    "startDate",
    "endDate",
    "floatStart",
    "floatEnd",
    "startDelay",
    "endDelay",
    "diffStart",
    "diffEnd",
    "compounding",
    "calculationId",
    "annualized",
    "annualDays",
    "benchmark",
    "riskfree",
    "windowType",
    "windowSize",
    "stepSize",
    "requireContinueData",
    "fit",
    "scalType",
    "scalValue",
    "scalPercentValue",
    "timehorizon",
]

HIDDEN_DATA_POINT_SETTINGS = [
    "type",
    "isFilterable",
    "isEpdp",
    "isCustomCalc",
    "canBeAddedToDataset",
    "mstarIpType",
    "nonstandardDisplayType",
    "hasOptions",
    "requireValueSearch",
    "scale",
    "reverseSign",
    "decimalPlaces",
    "displayOptions",
    "align",
    "width",
    "cellType",
    "sortLocal",
    "displayType",
    "percentileRank",
    "hide",
    "custom",
    "fixFormat",
    "customType",
    "generalFrequency",
    "calcCurType",
    "calcSdType",
    "calcUse5Days",
    "calcS1",
    "calcS2",
    "calcCfgDataSource",
    "calcBestReturnNum",
    "calcWorstReturnNum",
    "calcRollingTimePeriod",
    "calcReinvestDateType",
    "calcReinvestSource",
    "calcAdditionalSource",
    "calcIsApplyIsraelsenModification",
    "calcDaysOfAnnuYear",
    "calcMnav",
    "showType",
    "transType",
    "searchAttributes",
    "additionalReqId",
    "isPortfolioCalc",
    "asOfDate",
    "universe",
]



import inspect
import sys
import traceback
from functools import wraps
from typing import Any, Callable, Dict

import typeguard

from .._utils import format_analytics_logs, get_log_flag
from ._exceptions import (
    ApiResponseException,
    BadRequestException,
    QueryLimitException,
    ValueErrorException,
)

typechecked: Callable = typeguard.typechecked


def not_null(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        for arg in args:
            if not arg:
                raise BadRequestException("Null or empty parameter is not allowed.")

        return func(*args, **kwargs)

    return wrapper


def error_handler(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (BadRequestException, ValueErrorException) as e:
            wrappers = [x.name for x in traceback.extract_stack() if x.name == "error_handler"]
            if len(wrappers) > 1:
                raise ValueErrorException(str(e)) from None
            else:
                sys.stderr.write(str(e))
        except ApiResponseException as e:
            if e.status_code == 404:
                sys.stderr.write("Specified data not Found. ")
            else:
                raise ValueErrorException(str(e)) from None
        except QueryLimitException as e:
            # Some exceptions need to be propagated back to the caller as-is
            raise
        except Exception as e:
            raise ValueErrorException(str(e)) from None

    return wrapper



from typing import Any, Dict, List

import simplejson as json
from pandas import DataFrame

from .._base import _logger
from . import _decorator
from ._api import _direct_api_request
from ._config import _Config
from ._exceptions import ValueErrorException

_config = _Config()


def _data_point_request_builder(data_point_details: list) -> List[Any]:
    if not data_point_details:
        raise ValueErrorException("No datapoint available.")
    url = f"{_config.data_point_service_url()}v1/datapoints/datapointrequestbuilder"
    response_json: List[Any] = _direct_api_request("POST", url, json.dumps(data_point_details, ignore_nan=True))

    return response_json


def _request_asset_flow_data_points() -> List[Any]:
    url = f"{_config.securitydata_service_url()}v1/assetflow/datapoints"
    response_json: List[Any] = _direct_api_request("GET", url)
    return response_json


def _get_asset_flow_data_points_by_ids(data_point_ids: list) -> list:
    response_json = _request_asset_flow_data_points()
    result = []
    if response_json and isinstance(response_json, list):
        data_point_id_settings_dict = {x.get("datapointId", "").strip(): x for x in response_json if x is not None}
        for data_point_id in data_point_ids:
            settings = data_point_id_settings_dict.get(data_point_id)
            if settings is not None:
                settings = settings.copy()
                settings["datapointName"] = settings.pop("name")
                settings["alias"] = settings["datapointName"]
                result.append(settings)
    return result


@_decorator.error_handler
def _get_data_point_details(params: list) -> List[Dict[Any, Any]]:
    url = f"{_config.data_point_service_url()}v1/datapoints/detail"
    response_json: List[Dict[Any, Any]] = _direct_api_request("POST", url, json.dumps(params, ignore_nan=True))
    return _filter_time_series_data_points(params, response_json)


def _filter_time_series_data_points(
    data_points: List[Dict[Any, Any]], data_points_detail: List[Dict[Any, Any]]
) -> List[Dict[Any, Any]]:
    for data_point in data_points:
        data_point_id = data_point.get("datapointId")
        compare_id = lambda x, id=data_point_id: x.get("datapointId") == id
        should_keep_dp = (
            lambda x, id=data_point_id: x.get("datapointId") != id or x.get("isTsdp") != False or x.get("canBeAddedToDataset")
        )

        data_points_detail_for_id = list(filter(compare_id, data_points_detail))
        if len(data_points_detail_for_id) > 1:
            data_points_detail = list(filter(should_keep_dp, data_points_detail))
    return data_points_detail


@_decorator.error_handler
def _get_all_universes() -> DataFrame:
    url = f"{_config.data_point_service_url()}v1/universes"
    response_json = _direct_api_request("GET", url)
    result = DataFrame(response_json)
    return result[["id", "name"]]



import simplejson as json
from pandas import DataFrame

from .._config import _Config
from .._config_key import ALL_ASSET_FLOW_DATA_POINTS
from .._data_objects import DataPoints, Investments
from ..asset_flow import _asset_flow_api_request
from ._common import _get_data_points, _get_investment_ids
from ._data import Column, InvestmentDataRequest, InvestmentDataResults

_config = _Config()


class AssetFlowProvider:
    @staticmethod
    def build_request(
        investment_object: Investments, data_point_object: DataPoints, display_name: bool = False
    ) -> InvestmentDataRequest:
        investment_id_list = _get_investment_ids(investment_object)
        data_points = _get_data_points(investment_object, data_point_object, display_name)
        asset_flow_data_points = _filter_data_points_by_id(data_points, filter_list=ALL_ASSET_FLOW_DATA_POINTS)
        return InvestmentDataRequest(investment_id_list, asset_flow_data_points)

    @staticmethod
    def run_request(req: InvestmentDataRequest) -> InvestmentDataResults:
        return _get_asset_flow_data(req.investment_ids, req.data_points)


def _filter_data_points_by_id(df: DataFrame, filter_list: list) -> DataFrame:
    return df[df["datapointId"].isin(filter_list)].reset_index().drop(["index"], axis=1)


def _get_asset_flow_data(investment_ids: list, asset_flow_data_point_settings: DataFrame) -> InvestmentDataResults:
    investment_results = InvestmentDataResults()
    if asset_flow_data_point_settings is not None and not asset_flow_data_point_settings.empty:
        if "marketId" not in asset_flow_data_point_settings.columns:
            asset_flow_data_point_settings["marketId"] = None
        data_point_list = asset_flow_data_point_settings.to_dict(orient="records")
        market_data_point_dict = _group_by_market(data_point_list)

        for market_id, data_points in market_data_point_dict.items():
            url = f"{_config.securitydata_service_url()}v1/assetflow/data"
            postbody = {"marketId": market_id, "datapoints": data_points}
            if investment_ids:
                postbody["investments"] = list(map(lambda x: {"id": _get_sec_id(x)}, investment_ids))
            sub_investment_results = _parse_raw_asset_flow_data_values(
                investment_ids,
                data_points,
                _asset_flow_api_request.do_post_request(url, json.dumps(postbody, ignore_nan=True)),
            )
            investment_results.merge_with(sub_investment_results, in_place=True)
    return investment_results


def _group_by_market(data_point_list: list) -> dict:
    market_data_point_dict: dict = dict()
    for data_point in data_point_list:
        market_id = data_point.get("marketId", None)
        if market_id is not None and len(market_id.strip()) > 0:
            market_data_points = market_data_point_dict.get(market_id, [])
            market_data_points.append(data_point)
            market_data_point_dict[market_id] = market_data_points
    return market_data_point_dict


def _parse_raw_asset_flow_data_values(investment_ids: list, data_points: list, response_json: list) -> InvestmentDataResults:
    sub_investment_results = InvestmentDataResults()
    if not response_json or not isinstance(response_json, list):
        return sub_investment_results
    sec_id_to_investment_id = {_get_sec_id(x): x for x in investment_ids}
    alias_name_dict = {x.get("alias", ""): x.get("datapointName", "") for x in data_points}
    data_points_dict = {x.get("alias", ""): {"datapointName": x.get("datapointName", ""), "isTsdp": True} for x in data_points}

    for investment in response_json:
        sec_id = investment.get("id", "")
        investment_id = sec_id_to_investment_id[sec_id]
        values = investment.get("values", [])
        for data_point_value in values:
            alias = data_point_value.get("alias", "").strip()
            value_list = data_point_value.get("value", [])
            column_data = [
                Column(name=_concat_str(alias_name_dict.get(alias, alias), x.get("date", "")), value=x.get("value", None))
                for x in value_list
            ]
            data_point_meta = data_points_dict.get(alias)
            sub_investment_results.add_column_data(investment_id, alias, column_data)
            sub_investment_results.add_meta_data(investment_id, alias, data_point_meta)
    return sub_investment_results


def _concat_str(s1: str, s2: str) -> str:
    return s1 + " - " + s2


def _get_sec_id(investment_id: str) -> str:
    assert isinstance(investment_id, str)
    sec_id, *_ = investment_id.split(";")
    return sec_id



import time
import warnings
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import requests
import simplejson as json
from pandas import DataFrame

from .._base import _logger
from . import _decorator, _error_messages
from ._config import _Config
from ._data_objects._data_points_object import DataPoints
from ._data_type import ErrorMessages
from ._exceptions import BadRequestException, ResourceNotFoundError, ValueErrorException
from ._investment._normal_data import _parse_raw_normal_data_values
from ._portfolio_data_set import (
    DataSetType,
    PortfolioDataSet,
    _get_data_points_default_settings,
    _get_data_set_with_id,
    _get_data_sets,
)
from ._utils import _get_iso_today
from .user_items.portfolio import PortfolioDataApiBackend, get_portfolios

_config = _Config()
_portfolio_api = PortfolioDataApiBackend()


@_decorator.error_handler
@_decorator.typechecked
def get_data_sets() -> DataFrame:
    """Returns all Morningstar-created portfolio data sets.

    :Returns:
        DataFrame: A DataFrame object with data sets. DataFrame columns include:

        * data_set_id
        * name

    :Examples:
        Get all portfolio data sets.

    ::

        import morningstar_data as md

        df = md.direct.portfolio.get_data_sets()
        df

    :Output:
        ===========    ===============
        data_set_id	   name
        ===========    ===============
        1              Snapshot
        2              Returns (Daily)
        ===========    ===============

    """
    all_view_df = pd.DataFrame({"data_set_id": [], "name": []})
    df = _get_data_sets()
    df = df.sort_values(by=["id"])
    all_view_df["data_set_id"] = df["id"].astype(str)
    all_view_df["name"] = df["name"]
    all_view_df = all_view_df.reset_index(drop=True)
    return all_view_df


@_decorator.typechecked
def get_data_set_data_points(data_set_id: str) -> DataFrame:
    """Returns all data points for the given portfolio data set ID.

    Args:
        data_set_id (:obj:`str`): The unique identifier of a portfolio data set, e.g., '1'.

    :Returns:
        DataFrame: A DataFrame object with data points. DataFrame columns include:

        * data_point_id
        * name

    :Examples:
        Get data points by data set ID.

    ::

        import morningstar_data as md

        df = md.direct.portfolio.get_data_set_data_points(data_set_id="1")
        df

    :Output:

        ==============   ===============
        data_point_id    name
        ==============   ===============
        OS01W		     Name
        LS05M		     Base Currency
        ...              ...
        ==============   ===============

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    view = _get_data_set_with_id(data_set_id)
    data_points = pd.DataFrame(view["datapoints"])
    if len(data_points):
        data_points = data_points[["datapointId", "alias"]]
        data_points = data_points.rename(columns={"alias": "name"})
        data_points = data_points.rename(columns={"datapointId": "data_point_id"})
        return data_points
    else:
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET_DATA_POINT) from None


def _get_data_points_by_view_id(id: str, currency: Optional[str]) -> List[Dict[Any, Any]]:
    view = _get_data_set_with_id(id)
    return _set_currency(view["datapoints"], currency)


def _set_currency(data_points: List[Dict], currency: Optional[str]) -> List[Dict[Any, Any]]:
    if currency is not None:
        for data_point in data_points:
            if data_point.get("currency") is not None:
                data_point["currency"] = currency
    return data_points


@_decorator.not_null
@_decorator.typechecked
def get(portfolio_type: Optional[str] = None) -> DataFrame:
    warnings.warn(
        "The portfolio.get function is deprecated and will be removed in the next major version. Use user_items.get_portfolios instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_portfolios(portfolio_type=portfolio_type)


@_decorator.typechecked
def get_holdings(
    portfolio_ids: List[str],
    date: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
) -> DataFrame:
    """Returns holdings for the given portfolio and date range.

    Args:
        portfolio_ids (:obj:`list`): List of portfolio IDs. The maximum number is 5. Use the `get_portfolios <./portfolio_list.html#morningstar_data.direct.user_items.get_portfolios>`_ function to discover saved portfolios.
        date (:obj:`str`, `optional`): Holdings date. When this value is provided, `start_date` and `end_date` parameters are ignored. Use the `get_holding_dates <#morningstar_data.direct.portfolio.get_holding_dates>`_ function to discover available holding dates.
        start_date (:obj:`str`, `optional`): Start date of the date range.
        end_date (:obj:`str`, `optional`): End date of the date range.

    :Returns:
        DataFrame: A DataFrame object with portfolio holdings data. DataFrame columns include:

        * ObjectId
        * Portfolio Date
        * HoldingId
        * Weight
        * Name
        * Currency
        * ISIN
        * Ticker

    :Examples:
        Get holdings by portfolio ID and date.

    ::

        import morningstar_data as md

        df = md.direct.portfolio.get_holdings(
            portfolio_ids=['de21828b-91b7-4fe1-b66e-760fd5e657bc;BM'], # Replace with valid portfolio ID
            date = "2017-09-30")
        df

    :Output:
        =======================================   ==============   ==============   =======   =================================   ============   ============   ==========
        ObjectId                                  Portfolio Date   HoldingId	    Weight	  Name                                Currency       ISIN           Ticker
        =======================================   ==============   ==============   =======   =================================   ============   ============   ==========
        de21828b-91b7-4fe1-b66e-760fd5e657bc;BM   2017-09-30	   FOUSA00DFS;F     100	      BlackRock Global Allocation Inv A	  USD	         US09251T1034   MDLOX
        ...
        =======================================   ==============   ==============   =======   =================================   ============   ============   ==========

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if not portfolio_ids:
        raise BadRequestException("Please input portfolio_ids to proceed with your query.")
    combined_holdings = pd.DataFrame(
        {
            "ObjectId": [],
            "Portfolio Date": [],
            "HoldingId": [],
            "Weight": [],
            "Name": [],
            "Currency": [],
            "ISIN": [],
            "Ticker": [],
        }
    )
    if portfolio_ids:
        if len(portfolio_ids) > 5:
            raise BadRequestException("The number of portfolio_ids per query are limited to 5.")
        portfolio_available_dates = _get_dates(portfolio_ids, date, start_date, end_date)
        if not portfolio_available_dates and date is not None and len(date.strip()) > 0:
            portfolio_available_dates = {x: [date] for x in portfolio_ids}
        _check_volume(portfolio_available_dates)
        holdings_df = _get_holdings(portfolio_available_dates)
        combined_holdings = pd.concat([combined_holdings, holdings_df])

    if combined_holdings.empty:
        if len(portfolio_ids) == 1:
            raise ResourceNotFoundError(
                (_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_HOLDING).format(portfolio_ids)
            ) from None
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_HOLDINGS) from None
    else:
        combined_holdings = combined_holdings.where(combined_holdings.notnull(), None)
        combined_holdings.reset_index(drop=True, inplace=True)
    return combined_holdings


def _get_holdings(portfolio_available_dates: dict) -> DataFrame:
    holdings_list = []
    for portfolio_id, dates in portfolio_available_dates.items():
        for date in dates:
            holdings = _get_holdings_by_id(portfolio_id, date)
            holdings = holdings if holdings else []
            holdings_list.extend(
                [
                    {
                        "ObjectId": portfolio_id,
                        "Portfolio Date": date,
                        "HoldingId": x.get("id", None),
                        "Weight": x.get("weight", None),
                        "Name": x.get("name", None),
                        "Currency": x.get("currency", None),
                        "ISIN": x.get("isin", None),
                        "Ticker": x.get("ticker", None),
                    }
                    for x in holdings
                ]
            )
    return DataFrame(holdings_list)


def _check_volume(portfolio_available_dates: dict) -> None:
    if portfolio_available_dates:
        total = 0
        for dates in portfolio_available_dates.values():
            total += len(dates) if dates is not None and isinstance(dates, list) else 0
        if total > 10:
            raise ValueErrorException(
                "Please reduce the number of investments or shorten the query time period to proceed your query without interfering by too large data size."
            )


def _get_holdings_by_id(portfolio_id: str, date: str) -> list:
    try:
        url = f"{_config.portfolio_service_url()}/portfoliodataservice/v1/portfolios/{portfolio_id}/holdings-summary?portfolioDate={date}"
        response_json: list = _portfolio_api.do_get_request(url)
        return response_json
    except Exception as portfolio_error:
        if portfolio_error.__class__.__name__ != "ResourceNotFoundError":
            raise portfolio_error from None
        pass
    return []


def _get_dates(
    portfolio_ids: list,
    date: Optional[str],
    start_date: Optional[str],
    end_date: Optional[str],
) -> dict:
    if date is not None and len(date.strip()) > 0:
        date = _format_date(date)
    elif start_date is not None and len(start_date.strip()) > 0:
        start_date = _format_date(start_date)
        if end_date is None or len(end_date.strip()) == 0:
            end_date = _get_iso_today()
        end_date = _format_date(end_date)
    elif end_date is not None and len(end_date.strip()) > 0:
        raise BadRequestException("Please specify a portfolio_date or time period to proceed with your query.")

    response_json = _get_multi_portfolio_dates(portfolio_ids)

    if date is not None and len(date.strip()) > 0:
        return _handle_single_date(response_json, date)
    elif start_date is not None and len(start_date.strip()) > 0 and end_date is not None and len(end_date.strip()) > 0:
        return _handle_date_range(response_json, start_date, end_date)
    else:
        return _handle_no_date(response_json)


def _get_multi_portfolio_dates(portfolio_ids: list) -> dict:
    portfolio_date_dict = dict()
    for id in portfolio_ids:
        if id:
            portfolio_date_dict[id] = _get_holding_dates(id)
    return portfolio_date_dict


def _handle_single_date(portfolio_dates: dict, input_portfolio_date: str) -> dict:
    portfolio_available_dates = dict()
    for portfolio_id, dates in portfolio_dates.items():
        available_date = _get_available_date(input_portfolio_date, dates)
        if available_date is not None:
            portfolio_available_dates[portfolio_id] = [available_date]
    return portfolio_available_dates


def _get_available_date(target_date: str, dates: list) -> Optional[str]:
    dates = dates if dates else []
    if target_date in dates:
        return target_date

    available_date = None
    dates.sort()
    dates.reverse()
    for date in dates:
        if date < target_date:
            available_date = date
            break
    return available_date


def _handle_no_date(portfolio_dates: dict) -> dict:
    portfolio_available_dates = dict()
    for portfolio_id, dates in portfolio_dates.items():
        dates = dates if dates else []
        date = next((x for x in dates if x is not None and len(x.strip()) > 0), None)
        if date is not None and len(date) > 0:
            portfolio_available_dates[portfolio_id] = [date]
    return portfolio_available_dates


def _handle_date_range(portfolio_dates: dict, start_date: str, end_date: str) -> dict:
    portfolio_available_dates = dict()
    for portfolio_id, dates in portfolio_dates.items():
        available_dates = list()
        dates = dates if dates else []
        for date in dates:
            if start_date <= date <= end_date:
                available_dates.append(date)
        if available_dates:
            portfolio_available_dates[portfolio_id] = available_dates

    return portfolio_available_dates


@_decorator.typechecked
def get_holding_dates(portfolio_ids: Optional[Union[List[str], str]] = None) -> DataFrame:
    """Returns portfolio holdings dates for the given portfolios.


    Args:
        portfolio_ids (:obj:`list`, `optional`): A portfolio ID or list of portfolio IDs. If no ID is specified, the function will return data for all portfolios saved in the user's account. Use the `get_portfolios <./portfolio_list.html#morningstar_data.direct.user_items.get_portfolios>`_ function to discover saved portfolios.

    :Returns:
        DataFrame: A DataFrame object with all portfolio holdings dates. DataFrame columns include:

        * portfolioId
        * date

    :Examples:
        Get portfolio holdings dates for the given portfolio ID.

    ::

        import morningstar_data as md

        df = md.direct.portfolio.get_holding_dates(
            portfolio_ids=['07c317df-a4a7-4297-afc3-0c18bb79a672;UA']) # Replace with a valid portfolio ID
        df

    :Output:
        =======================================        ==========
        PortfolioId                                    Date
        =======================================        ==========
        07c317df-a4a7-4297-afc3-0c18bb79a672;UA        2021-10-01
        07c317df-a4a7-4297-afc3-0c18bb79a672;UA        2021-09-01
        ...
        =======================================        ==========

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if not isinstance(portfolio_ids, list) and portfolio_ids is not None:
        single_id = portfolio_ids
        portfolio_ids = list()
        portfolio_ids.append(single_id)
    portfolio_date_list = list()
    if not portfolio_ids:
        portfolio_df = get()
        portfolio_ids = list(portfolio_df["PortfolioId"])
    for portfolio_id in sorted(set(portfolio_ids), key=portfolio_ids.index):
        if portfolio_id is None or len(portfolio_id.strip()) == 0:
            continue
        dates = _get_holding_dates(portfolio_id, portfolio_ids)
        if not dates:
            continue
        portfolio_date_list.extend([{"PortfolioId": portfolio_id, "Date": date} for date in dates])
    portfolio_df = DataFrame(portfolio_date_list)
    if portfolio_df.empty:
        if len(portfolio_ids) == 1:
            raise ResourceNotFoundError((_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ID).format(portfolio_id)) from None
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_IDS) from None
    return portfolio_df


def _get_holding_dates(portfolio_id: str, portfolio_ids: Optional[Union[List[str], str]] = None) -> list:
    try:
        url = f"{_config.portfolio_service_url()}/portfoliodataservice/v1/portfolios/{portfolio_id}/dates"
        response_json: Dict[str, list] = _portfolio_api.do_get_request(url)
        return response_json["dates"]
    except Exception as portfolio_error:
        if portfolio_error.__class__.__name__ != "ResourceNotFoundError":
            raise portfolio_error from None
        pass
    return []


def _format_date(date: str) -> str:
    try:
        return time.strftime("%Y-%m-%d", time.strptime(date, "%Y-%m-%d"))
    except Exception as e:
        # _logger.error(f"Date format error: {e}")
        raise BadRequestException(ErrorMessages.date_format_error.value + f" Date format error: {e}.") from None


@_decorator.typechecked
def get_data(
    portfolio_id: str,
    data_set_id: Optional[Union[str, PortfolioDataSet, None]] = None,
    currency: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    data_point_settings: Optional[DataFrame] = None,
) -> DataFrame:
    """Returns data for the given portfolio and data points.

    Args:
        portfolio_id (:obj:`str` ): Portfolio ID. Use the `get_portfolios <./portfolio_list.html#morningstar_data.direct.user_items.get_portfolios>`_ function to discover saved portfolios.
        data_set_id (:obj:`str`, `optional`): Saved portfolio data set, e.g., "1". Use the `get_data_sets <./data_set.html#morningstar_data.direct.portfolio.get_data_sets>`_ function to discover available data sets.
        currency (:obj:`str`, `optional`): Currency setting for the data point values, when applicable. Use the `currency_codes <../morningstar_data/lookup.html#morningstar_data.lookup.currency_codes>`_ function to discover available currency codes.
        start_date (:obj:`str`, `optional`): Start date of a date range for retrieving data.
            The format is YYYY-MM-DD, e.g., '2020-01-01'.
        end_date (:obj:`str`, `optional`): End date of a date range for retrieving data. If no value is provided for `end_date`, current date will be used.
            The format is YYYY-MM-DD, e.g., '2020-01-01'.
        data_point_settings (:obj:`DataFrame`, `optional`): A DataFrame of data points with all defined settings. Each row represents a data point.
            Each column is a configurable setting. Users can get this DataFrame by using `get_data_set_data_points <./data_set.html#morningstar_data.direct.portfolio.get_data_set_data_points>`_ or `get_data_point_settings <../morningstar_data/lookup.html#morningstar_data.direct.get_data_point_settings>`_ for a given data point ID(s).
            Users can update setting values in this DataFrame if desired. Additionally, the priority of currency/start_date/end_date in `data_point_settings`
            is lower than the currency/start_date/end_date function parameters.


    :Returns:
        DataFrame: A DataFrame object with portfolio data.

    :Examples:
        Get portfolio data for the given portfolio and data set.

    ::

        import morningstar_data as md

        df = md.direct.portfolio.get_data(
            portfolio_id="8c6de08f-a668-4e78-a688-3d809aeb29b7;UA", # Replace with a valid portfolio ID
            data_set_id="1")
        df

    :Output:
        =========================================  =====   ==============   ================   =====   =================================   ==================================
        Id                                         Name    Base Currency    Portfolio Date     ...     Total Ret Annlzd 10 Yr (Year-End)   Total Ret Annlzd 15 Yr (Year-End)
        =========================================  =====   ==============   ================   =====   =================================   ==================================
        8c6de08f-a668-4e78-a688-3d809aeb29b7;UA    test1   US Dollar        2022-02-28         ...     None                                None
        ...
        =========================================  =====   ==============   ================   =====   =================================   ==================================

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    all_data = _get_portfolio_data(
        portfolio_id=portfolio_id,
        data_set_id=data_set_id,
        currency=currency,
        start_date=start_date,
        end_date=end_date,
        data_point_settings=data_point_settings,
    )
    col_name = all_data.columns.tolist()
    all_data = all_data.reindex(columns=col_name)
    return all_data


def _get_portfolio_data(
    portfolio_id: str,
    data_set_id: Optional[Union[str, PortfolioDataSet, None]] = None,
    currency: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    data_point_settings: Optional[DataFrame] = None,
) -> DataFrame:
    """
    As the "get_portfolio_data" API may been called by other APIs, so we extracted it as a private method.
    """
    # try:
    # build columns
    settings_data_frame = None
    if data_point_settings is not None and not data_point_settings.empty:
        settings_data_frame = _get_data_settings_data_points(data_point_settings, currency, start_date, end_date)
    else:
        if isinstance(data_set_id, PortfolioDataSet):
            data_set_id = data_set_id.data_set_id
            if get_data_set_data_points(data_set_id=data_set_id).empty:
                raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET_DATA_POINT) from None
        if not data_set_id:
            raise BadRequestException("Please specify either data_set_id or data_point_settings to proceed with your query.")
        settings_data_frame = DataFrame(_get_data_set_data_points(data_set_id, currency, start_date, end_date))
    portfolio_ids = [portfolio_id]

    settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)

    # get data
    raw_data = _get_normal_data(
        portfolio_ids,
        _filter_disable_data_points(settings_data_frame).to_dict(orient="records"),
    )

    investment_results = _parse_raw_normal_data_values(raw_data)

    # add id
    all_data = investment_results.as_data_frame()
    col_name = all_data.columns.tolist()
    all_data = all_data.reindex(columns=col_name)

    return all_data


def _filter_disable_data_points(data_point_settings: DataFrame = None) -> DataFrame:
    columns = data_point_settings.columns
    if "disable" in columns:
        return data_point_settings.loc[data_point_settings["disable"] != True]
    return data_point_settings


def _get_normal_data(portfolio_ids: list, data_points: list) -> DataFrame:
    if not portfolio_ids:
        raise BadRequestException("There is no portfolio data returned.")
    postbody = {
        "datapoints": data_points,
        "investments": list(map(lambda x: {"id": x}, portfolio_ids)),
    }
    resp: DataFrame = _request_porfolio_data(postbody, portfolio_ids)
    if resp is None:
        raise BadRequestException("There is no return data available.")
    return resp


def _request_porfolio_data(postbody: dict, portfolio_ids: list) -> DataFrame:
    job_detail, portfolio_data = _get_portfolio_job(postbody)
    job_id = job_detail["id"]
    try_times = 10
    while True:
        job_detail, portfolio_data = _get_portfolio_job(postbody, job_id)
        time.sleep(2)
        try_times = try_times - 1
        if job_detail.get("percentage", "0") == "100" or try_times == 0:
            break
    if not portfolio_data:
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_GET_DATA_RETURN.format(portfolio_ids[0]))
    return portfolio_data


def _get_portfolio_job(postbody: dict, job_id: Optional[str] = None) -> tuple:
    url = f"{_config.portfolio_service_url()}/portfoliodataservice/v1/portfolios/calculation?maxExecTime=600000&maxWaitTime=600000&timeOut=600000"
    if job_id:
        url = url + f"&jobId={job_id}"
    response_json = _portfolio_api.do_post_request(url, json.dumps(postbody, ignore_nan=True).replace("NaN", "null"))
    job_detail = response_json["job"]
    portfolio_data = None

    if job_detail.get("percentage", "0") == "100":
        portfolio_data = response_json
    return job_detail, portfolio_data


def _get_data_settings_data_points(
    data_point_settings: DataFrame,
    currency: Optional[str],
    start_date: Optional[str],
    end_date: Optional[str],
) -> DataFrame:
    if "datapointId" not in data_point_settings.columns:
        raise BadRequestException("Please input valid data_point_settings to proceed with your query.")
    data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
    start_date, end_date = _format_start_end_date(start_date, end_date)
    # only set currency to the data_points with currency.
    # only set start_date/end_date to the datapoints without isTsdp=False.

    data_point_default_settings = _get_data_points_default_settings(data_point_settings["datapointId"].tolist())

    settings_list = data_point_settings.to_dict(orient="records")
    for settings in settings_list:
        data_point_id = settings["datapointId"]
        defined_data_point = data_point_default_settings.get(data_point_id)
        if defined_data_point is not None and defined_data_point.get("currency") is None:
            settings["currency"] = None
        elif currency is not None and len(currency.strip()) > 0:
            settings["currency"] = currency
        if defined_data_point is not None and defined_data_point.get("isTsdp") is not None and settings.get("isTsdp") is None:
            settings["isTsdp"] = defined_data_point.get("isTsdp")

        if (
            settings.get("isTsdp") is not False
            and start_date is not None
            and len(start_date.strip()) > 0
            and end_date is not None
            and len(end_date.strip()) > 0
        ):
            settings["startDate"] = start_date
            settings["endDate"] = end_date

        esg_data_point = data_point_default_settings.get(data_point_id)
        if data_point_id != "OS01W" and esg_data_point is not None and settings.get("isTsdp") is None:
            if esg_data_point.get("isTsdp") is not None:
                settings["isTsdp"] = esg_data_point.get("isTsdp")
            else:
                start = settings.get("startDate")
                end = settings.get("endDate")
                if start is not None and len(start.strip()) > 0 and end is not None and len(end.strip()) > 0:
                    settings["isTsdp"] = True
                else:
                    settings["isTsdp"] = False
    data_point_settings = DataFrame(settings_list)
    data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
    data_points = DataPoints(data_point_settings)
    settings_data_frame = DataFrame(data_points.get_data_points().to_dict(orient="records"))
    settings_data_frame = _disable_data_points(settings_data_frame, data_point_default_settings)

    return settings_data_frame


def _disable_data_points(settings_data_frame: DataFrame, defined_data_points: dict) -> DataFrame:
    settings_list = settings_data_frame.to_dict(orient="records")
    for settings in settings_list:
        data_point_id = settings["datapointId"]
        frequency_m_data_point = defined_data_points.get(data_point_id)
        if frequency_m_data_point is not None:
            frequency = settings.get("frequency")
            defined_freq = frequency_m_data_point.get("frequency", None)
            if defined_freq is not None and frequency is not None and len(frequency.strip()) > 0 and frequency != defined_freq:
                settings["disable"] = True

        defined_data_point = defined_data_points.get(data_point_id)
        # for defined datapoint with isTsdp = False
        if defined_data_point is not None:
            if defined_data_point.get("isTsdp") is False and settings.get("isTsdp") is True:
                settings["disable"] = True
            if defined_data_point.get("isTsdp") is True and settings.get("isTsdp") is False:
                settings["disable"] = True
    settings_data_frame = DataFrame(settings_list)
    settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)
    return settings_data_frame


def _get_data_set_data_points(
    data_set_id: str,
    currency: Optional[str],
    start_date: Optional[str],
    end_date: Optional[str],
) -> list:
    data_point_ids = _get_data_points_by_view_id(data_set_id, currency)
    data_points = DataPoints(data_point_ids)
    if not data_point_ids:
        return list(data_points.get_data_points().to_dict(orient="records"))
    data_set_type = PortfolioDataSet.get_data_set_type(data_set_id)
    if data_set_type == DataSetType.TimeSeries:
        if start_date is None or len(start_date.strip()) == 0:
            raise BadRequestException(_error_messages.BAD_REQUEST_ERROR_NO_START_AND_END_DATE) from None
        if end_date is None or len(end_date.strip()) == 0:
            raise BadRequestException(_error_messages.BAD_REQUEST_ERROR_NO_END_DATE)
        for data_point_id in data_point_ids:
            if data_point_id.get("isTsdp") is True:
                data_point_id["startDate"] = start_date
                data_point_id["endDate"] = end_date

    if data_set_type == DataSetType.Custom_Calculation:
        start_date, end_date = _format_start_end_date(start_date, end_date)
        if start_date is not None and len(start_date.strip()) > 0 and end_date is not None and len(end_date.strip()) > 0:
            for data_point_id in data_point_ids:
                data_point_id["startDate"] = start_date
                data_point_id["endDate"] = end_date
        return list(data_points.get_data_points().to_dict(orient="records"))

    if data_set_type == DataSetType.Current_Or_TimeSeries:
        start_date, end_date = _format_start_end_date(start_date, end_date)
        for data_point_id in data_point_ids:
            if (
                data_point_id.get("isTsdp") is not False
                and start_date is not None
                and len(start_date.strip()) > 0
                and end_date is not None
                and len(end_date.strip()) > 0
            ):
                data_point_id["startDate"] = start_date
                data_point_id["endDate"] = end_date
                data_point_id["isTsdp"] = True
                # support 'm' only
                data_point_id["frequency"] = "m"
            else:
                data_point_id["isTsdp"] = False
        return list(data_points.get_data_points().to_dict(orient="records"))

    return data_point_ids


def _format_start_end_date(start_date: Optional[str], end_date: Optional[str]) -> tuple:
    if start_date is not None and len(start_date.strip()) > 0:
        start_date = _format_date(start_date)
        if end_date is None or len(end_date.strip()) == 0:
            raise BadRequestException(ErrorMessages.start_end_date_error.value)
        end_date = _format_date(end_date)
    elif end_date is not None and len(end_date.strip()) > 0:
        raise BadRequestException(ErrorMessages.start_end_date_error.value)
    return start_date, end_date


@_decorator.typechecked
def get_benchmark_settings(portfolio_id: str) -> DataFrame:
    """
    Returns a portfolio's benchmark settings.

    Args:
        portfolio_id (:obj:`str` ): Portfolio ID. Use the `get_portfolios <./portfolio_list.html#morningstar_data.direct.user_items.get_portfolios>`_ function to discover saved portfolios.

    :Returns:
        DataFrame: A DataFrame object containing the portfolio settings.

    :Examples:

    Get portfolio settings for a custom model portfolio.
    ::
        import morningstar_data as md

        df = md.direct.portfolio.get_settings(
            portfolio_id="d38b96dd-fbe1-4750-b05d-fa7cf4b6e35a;MD") # Replace with a valid portfolio ID
        df

    :Output:

        ========================================  ====================================  ====================================  ================  ======================  ==============  ================  ======================  =====================
        Portfolio ID                              Risk Free Proxy ID                    Benchmark 1 ID                        Benchmark 1 Name  Benchmark 1 SecurityId  Benchmark 2 ID  Benchmark 2 Name  Benchmark 2 SecurityId  Internal Portfolio ID
        ========================================  ====================================  ====================================  ================  ======================  ==============  ================  ======================  =====================
        d38b96dd-fbe1-4750-b05d-fa7cf4b6e35a;MD   e9b23e21-9c6f-4696-9583-3b3729b4ad99  fcff80f8-2fbc-4211-bbce-45252008a9cc  S&P 500 TR USD    XIUSA04G92;XI           None            None              None                    None

        ========================================  ====================================  ====================================  ================  ======================  ==============  ================  ======================  =====================

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """

    _logger.info("Fetching portfolio settings by id")
    portfolio_settings = _get_settings_by_id(portfolio_id=portfolio_id)
    settings = {}
    settings["Porfolio ID"] = portfolio_settings["portfolio"]["portfolioId"]
    settings["Internal Portfolio ID"] = portfolio_settings["portfolio"]["portfolio2Id"]
    settings["Risk Free Proxy ID"] = portfolio_settings["portfolio"]["riskFreeProxyId"]

    primary_benchmark = portfolio_settings["portfolio"]["benchmarkId"]
    secondary_benchmark = portfolio_settings["portfolio"]["secondaryBenchmarkId"]
    # The secondary benchmark is optional so not all portfolios will have this information
    secondary_benchmark_name = None
    secondary_benchmark_id = None

    for benchmark in portfolio_settings["benchmarks"]:
        benchmark_id = benchmark["benchmarkId"]
        if benchmark_id == primary_benchmark:
            primary_benchmark_name = benchmark["name"]
            primary_benchmark_sec_id = benchmark["securityId"]

        elif benchmark_id == secondary_benchmark:
            secondary_benchmark_name = benchmark["name"]
            secondary_benchmark_id = benchmark["securityId"]

    settings["Benchmark 1 ID"] = primary_benchmark
    settings["Benchmark 1 Name"] = primary_benchmark_name
    settings["Benchmark 1 Security ID"] = primary_benchmark_sec_id

    settings["Benchmark 2 ID"] = secondary_benchmark
    settings["Benchmark 2 Name"] = secondary_benchmark_name
    settings["Benchmark 2 Security ID"] = secondary_benchmark_id

    return DataFrame(settings, index=[0])


def _get_settings_by_id(portfolio_id: str) -> dict:
    try:
        url = f"{_config.portfolio_service_url()}/portfoliodataservice/v1/portfolios/{portfolio_id}/settings"
        response_json = _portfolio_api.do_get_request(url)
        settings: dict = response_json["settings"]  # settings key is always present when 200 status code is returned.
        return settings
    except Exception as portfolio_error:
        if portfolio_error.__class__.__name__ != "ResourceNotFoundError":
            raise portfolio_error from None
        pass
    return {}



import enum
from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from ._exceptions import BadRequestException


class DatabaseCD(str, Enum):
    user = ("UDP",)
    firm = ("CDP",)


class RaiseEnum(enum.EnumMeta):
    def __getitem__(cls, name: str) -> Any:
        try:
            return super().__getitem__(name)
        except KeyError:
            options = ", ".join(cls._member_map_.keys())
            raise BadRequestException(
                f"Please choose one of the following valid options: '{options}'.\n" f"'{name}' is an invalid choice.\n"
            ) from None


class ErrorMessages(str, Enum):
    date_format_error = "Please input the date in yyyy-MM-dd format."
    start_date_require_error = "Please specify start_date to proceed with your query."
    no_data_point_msg = "There is no data point available."
    start_end_date_error = "Please specify start_date and end_date to proceed with your query."
    frequency_error_msg = "Please specify frequency to proceed with your query."
    currency_error_msg = "Please specify currency to proceed with your query."


@dataclass()
class DryRunResults:
    estimated_cells_used: int
    daily_cells_remaining_before: int
    daily_cell_limit: int
    daily_cells_remaining_after: int = field(init=False)

    def __post_init__(self) -> None:
        self.daily_cells_remaining_after = self.daily_cells_remaining_before - self.estimated_cells_used



from __future__ import annotations

from typing import Any, Dict, List, Optional

import simplejson as json
from pandas import DataFrame

from ..._base import _logger
from .._api import _direct_api_request
from .._config import _Config
from .._error_messages import (
    BAD_REQUEST_ERROR_NO_PORTFOLIO_HOLDING_ID,
    BAD_REQUEST_INVALID_PORTFOLIO_HOLDING_DATA_FRAME,
    BAD_REQUEST_PORTFOLIO_WEIGHT_ERROR,
)
from .._exceptions import BadRequestException
from .._utils import _format_date
from ._common import PortfolioDataApiBackend, PortfolioJSONEncoder

_portfolio_api = PortfolioDataApiBackend()
_config = _Config()


class Holding:
    holding_id: str
    weight: Optional[float]
    market_value: float
    quantity: float
    price: float

    def __init__(self, holding_id: str, weight: Optional[float] = None) -> None:
        self.holding_id = holding_id
        self.weight = weight


class HoldingsEntity:
    holdings: List[Holding]
    portfolio_date: str
    total_amount: float = 10000

    def __init__(self, holdings: List, portfolio_date: str, total_amount: float = 10000) -> None:
        self.holdings = holdings
        self.total_amount = total_amount
        self.portfolio_date = _format_date(portfolio_date)

        self._validate_holdings_id_and_weight()
        self._validate_holdings_total_weight()

    def _validate_holdings_id_and_weight(self) -> None:
        """
        Validate portfolio holdings must have id and weight.
        """
        all_weight_is_none = True
        _logger.info(f"Validating portfolio holdings id and weight")
        for holding in self.holdings:
            if not holding.holding_id:
                # Find out holding without holding id.This is an invalid case.
                _logger.debug(f"The portfolio date {self.portfolio_date} holdings Holding_Id is missing")
                raise BadRequestException(BAD_REQUEST_ERROR_NO_PORTFOLIO_HOLDING_ID) from None
            if holding.weight is not None:
                all_weight_is_none = False

        if all_weight_is_none:
            # All holdings without weight.This is an valid case.
            _logger.info(
                f"The portfolio date {self.portfolio_date} all holdings weight is none.divide these holding weight by 100."
            )
            new_weight = 100 / len(self.holdings)
            for holding in self.holdings:
                holding.weight = new_weight
            _logger.debug(
                f"The portfolio date {self.portfolio_date} all holdings weight is None,set holdings weight: {new_weight}"
            )

    def _validate_holdings_total_weight(self) -> None:
        """
        Validation portfolio holdings total weight must equal 100.0 .
        """
        total_weight_error_msg = BAD_REQUEST_PORTFOLIO_WEIGHT_ERROR

        total_weight = sum(c.weight or 0 for c in self.holdings)  # get total weight.
        _logger.info(f"Validating the portfolio date {self.portfolio_date} holdings total weight: {total_weight}%.")
        if round(total_weight, 2) > 100:
            # Holdings total_weight>100.This is an invalid case.
            error_message = (
                f"The portfolio date: {self.portfolio_date}. {total_weight_error_msg}subtract {round(total_weight - 100, 2)}%."
            )
            _logger.debug(error_message)
            raise BadRequestException(error_message) from None
        if round(total_weight, 2) < 100:
            # Holdings total_weight<100.This is an invalid case.
            error_message = (
                f"The portfolio date: {self.portfolio_date}. {total_weight_error_msg}add {round(100 - total_weight, 2)}%."
            )
            _logger.debug(error_message)
            raise BadRequestException(error_message) from None

        for holding in self.holdings:  # Set holding market_value with weight.
            if holding.weight is not None:
                _logger.info(f"# Set holding market_value with weight.")
                holding.market_value = self.total_amount * holding.weight / 100
        return None

    def __str__(self) -> str:
        return f"{json.dumps(self.to_json(), ignore_nan=True, cls=PortfolioJSONEncoder, ensure_ascii=False)}"

    def to_json(self) -> Dict:
        # Set holding price and quantity
        self._set_holding_price_quantity()
        dict_holdings = []
        _logger.info(f"Generate holdings entity request json")
        for holding in self.holdings:
            dict_holdings.append(
                {
                    "secId": holding.holding_id,
                    "weight": holding.weight,
                    "marketValue": holding.market_value,
                    "quantity": holding.quantity,
                    "price": holding.price,
                }
            )
        to_return = {"portfolioDate": self.portfolio_date, "holdings": dict_holdings}
        return to_return

    def _set_holding_price_quantity(self) -> None:
        """
        Get and set portfolio holdings price and quantity .
        """
        _logger.info("Get and set the portfolio date price for holdings.")
        post_body = {
            "priceDate": self.portfolio_date,
            "investments": list(map(lambda x: {"id": x.holding_id}, self.holdings)),
        }
        _logger.debug(f"Get portfolio date price post body..{post_body}")
        url = f"{_config.securitydata_service_url()}/v1/securities/price"
        response_json = _portfolio_api.do_post_request(url, json.dumps(post_body, ignore_nan=True))
        _logger.debug(f"The portfolio date price for holdings..{response_json}")

        _logger.info("Set price and quantity for each holding.")
        for holding in self.holdings:
            # Get holding details from response by hoding_id.
            holding_info = [item for item in response_json if item.get("id") == holding.holding_id]
            price = float(holding_info[0].get("displayprice", 1))
            holding.price = price
            # get quantity by price.
            holding.quantity = holding.market_value / price
            _logger.debug(f"The {holding.holding_id} price:{price},quantity:{holding.quantity}")


class HoldingsEntityRequestBuilder:
    portfolio_id: Optional[str]

    def __init__(self, holdings_data_frame: DataFrame) -> None:
        self.holdings_entitys = self._data_frame_to_holdings_entitys(holdings_data_frame)

    def _data_frame_to_holdings_entitys(self, holdings_data_frame: DataFrame) -> List[HoldingsEntity]:
        """
        Convert portfolio holdings dataframe to holdings entitys.
        """
        _logger.info("Portfolio holdings dataframe to holdings entitys.")
        if not set(["Portfolio Date", "HoldingId", "Weight"]).issubset(holdings_data_frame.columns):
            # Holdings DataFrame columns must have Portfolio Date ,HoldingId, Weight.This is an invalid case.
            _logger.debug(
                f"Holdings data frame is missing required columns. It only has :{','.join(holdings_data_frame.columns)}"
            )
            raise BadRequestException(BAD_REQUEST_INVALID_PORTFOLIO_HOLDING_DATA_FRAME)

        # Group by Portfolio Date
        portfolio_dates = holdings_data_frame.groupby("Portfolio Date")
        # holdings dataframe to holdings entitys.
        holdings_entitys = []
        for portfolio_date, rows in portfolio_dates:
            holdings_entitys.append(
                HoldingsEntity(
                    portfolio_date=portfolio_date,
                    holdings=[(Holding(holding_id=row.HoldingId, weight=row.Weight)) for index, row in rows.iterrows()],
                )
            )
        return holdings_entitys

    def update_portfolio_holdings(self) -> Any:
        """
        Update portfolio holdings with holdings entitys.
        """
        _logger.info("Saving holdings to portfolio.")
        _logger.debug(f"Portfolio id: {self.portfolio_id}")
        url = f"{_config.portfolio_service_url()}portfoliodataservice/v1/portfolios/{self.portfolio_id}/holdings"
        response_json = _portfolio_api.do_post_request(
            url, json.dumps(self.holdings_entitys, cls=PortfolioJSONEncoder, ignore_nan=True)
        )
        _logger.debug("Update portfolio holdings response:{response}")
        return response_json



ACCESS_DENIED_ERROR = "Unauthorized request! Please login with valid credentials. For more information, please contact Analytics Lab support team at Morningstar."
FORBIDDEN_ERROR = "Forbidden request! You do not have permission to access this resource. For more information, please contact Analytics Lab support team at Morningstar."
NETWORK_ERROR = "Network connection error. Please check your internet connection."
TIMEOUT_ERROR = "Request timed out. Please try again or contact support at morningstardirect@moringstar.com"
BAD_REQUEST_ERROR = "Bad request! Server could not process your request at this time. This error can occur due to multiple reasons (for example; bad request syntax, invalid message structure/framing, or incorrect/corrupt request).\nPlease refer morningstar data package documentation to make a valid request: https://docs-analyticslab.morningstar.com/latest/morningstar_data.html"
CLIENT_ERROR = "Bad request! Server could not process your request at this time. This error can occur due to multiple reasons (for example; bad request syntax, invalid message structure/framing, or incorrect/corrupt request).\nPlease refer morningstar data package documentation to make a valid request: https://docs-analyticslab.morningstar.com/latest/morningstar_data.html"
INTERNAL_SERVER_ERROR = "Something went wrong on our side!  For more details, please contact Analytics Lab support team at Morningstar. We apologize for any inconvenience this may have caused"

# Add query limit error messages here
QUERY_LIMIT_ERROR = "You have reached your daily query limit. Please fill out this form for further assistance: https://go.morningstar.com/LP=6390"
QUERY_LIMIT_ERROR_SHOW_LIMIT = "You have reached your daily query limit of $query cells. Please fill out this form for further assistance: https://go.morningstar.com/LP=6808"

# AMS Flagged Features
DELIVERY_ACCESS_ERROR = (
    "You don't have access to this feature. Please fill out this form for further assistance: https://go.morningstar.com/LP=6809"
)

# Add custom bad request error messages here
BAD_REQUEST_ERROR_NO_START_AND_END_DATE = "Specify `start_date` and `end_date` to proceed with your query."
BAD_REQUEST_ERROR_NO_END_DATE = "Specify the `end_date` to proceed with your query."
BAD_REQUEST_ERROR_NO_START_DATE = "Specify the `start_date` to proceed with your query."
BAD_REQUEST_ERROR_INCLUDE_ALL_DATE = "Specify either date or both start date and end date to proceed with your query."
BAD_REQUEST_ERROR_INVALID_DATE_FORMAT = "Specify date in the format yyyy-MM-dd to proceed with your query."
BAD_REQUEST_ERROR_NO_INVESTMENT_IDS = "Specify the `investments` to proceed with your query."
BAD_REQUEST_ERROR_NO_INVESTMENT_IDS_FOR_DRY_RUN = "Specify the `investments` when using dry_run parameter."
BAD_REQUEST_ERROR_NO_PORTFOLIO_DATA = "No matching portfolio data for the given date or date range."
BAD_REQUEST_ERROR_INVALID_INVESTMENT_ID = "One or more ID's, in the provided list of `investments`, are invalid!"
BAD_REQUEST_ERROR_INVALID_INVESTMENT_LIST_ID = (
    "Specify the valid `list_id` to proceed with your query. A `list_id` is a global unique identifier i.e. uuid."
)
BAD_REQUEST_ERROR_INVALID_PORTFOLIO_ID = (
    "Specify the valid `portfolio_id` to proceed with your query. A `portfolio_id` is a global unique identifier i.e. uuid."
)
BAD_REQUEST_ERROR_INVALID_LOOKTHROUGH_HOLDING_TYPE = "Indexes are not supported as a valid Look-Through holding."
BAD_REQUEST_ERROR_NO_START_AND_END_DATE = "Please specify `start_date` and `end_date` to proceed with your query."
BAD_REQUEST_ERROR_NO_END_DATE = "Please specify the `end_date` to proceed with your query."
BAD_REQUEST_ERROR_DATA_SET = "Please enter a valid format for `data_set_id`. To retrieve your data sets, execute `md.direct.user_items.get_data_sets()` and select a valid `data_set_id`. To retrieve morningstar data sets, execute `md.direct.lookup.get_morningstar_data_sets()` "
BAD_REQUEST_ERROR_INVALID_DELIVER_CONFIG = "Delivery config is not valid. Check the docs for examples of accepted inputs."

# Add custom resource not found error messages here
RESOURCE_NOT_FOUND_ERROR = (
    "Requested resource does not exist. Please make sure that the provided input belongs to a resource in Direct."
)
RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA = "Requested resource not found with the provided `search_criteria_id`. Please make sure that the provided `search_criteria_id` is valid and exists. \nTo view all the available search criterias, execute - `md.direct.user_items.get_search_criteria()`."
RESOURCE_NOT_FOUND_ERROR_NO_SEARCH_CRITERIA = "Requested resource not found. Please make sure that you have search criterias available in your Direct account. \nTo view search criterias, log in to Direct and navigate to `Workspace >> Search Criteria >> My Search Criteria`.\nSearch tutorial: https://morningstardirect.morningstar.com/clientcomm/Searches.pdf"
RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA_ID = "Requested resource not found for search_criteria_id `{}`. Please make sure that provided search_criteria_id exists in your Direct account. To get all the search criterias, execute `md.direct.user_items.get_search_criteria()`"
RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA_NO_INVESTMENTS = "Could not find any investments in the provided `search criteria`. Please make sure that provided `search criteria` is valid. To get all the search criterias, execute `md.direct.user_items.get_search_criteria()`"

RESOURCE_NOT_FOUND_ERROR_NO_DATA_SETS = "No data sets found. To create a new data set, please open Direct and navigate to Workspace -> Data Sets -> My Data Sets -> New Data Set."
RESOURCE_NOT_FOUND_ERROR_DATA_SET = "The requested `data_set_id` does not exist. To retrieve your data sets, execute `md.direct.user_items.get_data_sets()` and select a valid `data_set_id`. To retrieve morningstar data sets, execute `md.direct.lookup.get_morningstar_data_sets()`"
RESOURCE_NOT_FOUND_ERROR_DATA_SET_DATA_POINT = "Please input a valid `data_set_id` to proceed with your query. To retrieve all the data sets, execute `md.direct.portfolio.get_data_sets()` and select a valid `data_set_id`."
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ACCOUNT = "No portfolios found for the given portfolio_type. \nTo create or see all of your existing portfolio accounts, log in to Direct and navigate to `Portfolio Management >> Accounts`. \nPortfolio management tutorial: https://morningstardirect.morningstar.com/clientcomm/Portfolios.pdf"
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ACCOUNTS = "Requested resources not found. Please make sure that you have created at least one portfolio account in Direct. \nTo create or see all of your existing portfolio accounts, log in to Direct and navigate to `Portfolio Management >> Accounts`. \nPortfolio management tutorial: https://morningstardirect.morningstar.com/clientcomm/Portfolios.pdf"
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_ID = "Requested resource not found for portfolio ID `{}`. Please make sure that a portfolio account exists in your Direct account with the provided ID. To get the holding dates for all your portfolio accounts, execute `md.direct.portfolio.get_holding_dates()`."
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_IDS = "Requested resources not found for provided portfolio IDs. Please make sure that portfolio accounts exist, for the provided list of IDs, in your Direct account. To get the holding dates for all your portfolio accounts, execute `md.direct.portfolio.get_holding_dates()`."
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_HOLDING = "Requested resource not found for portfolio ID `{}` and dates. Please make sure that, in Direct, a portfolio account exists with the provided ID and has holdings during provided dates. To get the holding dates for all your portfolio accounts, execute `md.direct.portfolio.get_holding_dates()`."
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_HOLDINGS = "Requested resources not found for portfolio IDs and dates. Please make sure that, in Direct, portfolio accounts exist with the provided IDs and have holdings during provided dates. To get the holding dates for all your portfolio accounts, execute `md.direct.portfolio.get_holding_dates()`."
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_GET_DATA_RETURN = (
    "Requested resource not found for portfolio ID `{}`. There is no return data available for the provided portfolio ID."
)

# Add custom resource not found error messages here for Asset Flow module
RESOURCE_NOT_FOUND_ASSET_FLOW = "The requested resource does not exist. Please make sure to provide a valid market_id."

# Add custom resource not found error messages here for Investments module
RESOURCE_NOT_FOUND_INVESTMENT_DATA = "No investments found for provided `investments` object. Please make sure that provided investments are valid and exist in your Direct account."
RESOURCE_NOT_FOUND_INVESTMENT_DATA_LIST_ID = "No investments found for the provided `list_id`. Please execute `md.direct.user_items.get_investment_lists()` to retrieve all the investment lists and select a valid `list_id`."
RESOURCE_NOT_FOUND_INVESTMENT_DATA_SEARCH_CRITERIA_ID = "No investments found for the provided `search_criteria_id`. Please execute `md.direct.user_items.get_search_criteria()` to retrieve all the available search criterias and select a valid `id`."
RESOURCE_NOT_FOUND_INVESTMENT_DATA_SEARCH_CRITERIA_CONDITION = "No investments found for the provided `search_criteria_condition`. Please execute `md.direct.user_items.get_search_criteria_conditions(search_criteria_id=:str)` with a valid search_criteria_id to build a search criteria condition."

# Add custom resource not found error messages here for Performance module
RESOURCE_NOT_FOUND_ERROR_PERFORMANCE_REPORT = (
    "The requested resource does not exist. Please make sure that the provided `report_id` is valid and exists."
)

# Add custom resource not found error messages here for Returns module
RESOURCE_NOT_FOUND_ERROR_NO_RETURNS_RETRIEVED = "Failed to get returns for provided `investments` argument."
RESOURCE_NOT_FOUND_ERROR_NO_RETURNS_RETRIEVED_FOR_INVESTMENT_LIST = "Could not fetch return for specified investment list. Please make sure that provided investments are valid and exist in your Direct account."
RESOURCE_NOT_FOUND_ERROR_NO_RETURNS_RETRIEVED_FOR_BENCHMARK_ID = "Could not fetch return for specified `benchmark_sec_id`. Please make sure that provided `benchmark_sec_id` is valid and exist in your Direct account."

# Add custom resource not found error messages here for Holding module
RESOURCE_NOT_FOUND_ERROR_HOLDING = "The requested resource does not exist. Make sure that the provided `investment_ids` are valid and have holdings within the specified date or date range."

# Add custom error messages here for Peer group module
BAD_REQUEST_ERROR_NO_DATA_POINT_ID_OR_ALIAS = "Specify datapointId and alias for each data_point."
BAD_REQUEST_ERROR_ALIAS_DUPLICATED = "The alias can't be duplicated."
BAD_REQUEST_ERROR_INCLUDE_NON_CALCULATION_DATA_POINTS = "Only calculated data points are supported."
BAD_REQUEST_ERROR_INCLUDE_DIFF_CALCULATION_DATA_POINTS = "Calculation is supported only on same data points."
BAD_REQUEST_ERROR_NO_DATA_POINTS = "No data_points."
BAD_REQUEST_ERROR_INVALID_CATEGORY_ID = "Specify category_id in the format 'category_id;universe' to proceed with your query."
BAD_REQUEST_ERROR_NO_INVESTMENT_SOURCE = "Specify the `investments` to proceed with your query."
BAD_REQUEST_ERROR_INVALID_PERCENTILES = "The percentiles should be integer if not null and range from 1 to 100."

# Add custom error messages here for investment list
RESOURCE_NOT_FOUND_ERROR_INVESTMENT_List = (
    "The requested resource does not exist. Please make sure that the provided `list_id` is valid and exists."
)

# Add custom error messages here for portfolio list
RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_List = "The requested resource does not exist. Make sure that the portfolios with provided `portfolio_type` is valid and exist in your Direct account."


# Add custom error messages here for Lookup
RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD = "Requested resource does not exist for the given `keyword`."
RESOURCE_NOT_FOUND_ERROR_LOOKUP_KEYWORD_UNIVERSE = "Requested resource does not exist for the given `keyword` and `universe`."


# Add custom error messages here for custom database
RESOURCE_NOT_FOUND_ERROR_CUSTOM_DATABASE = "Requested resource does not exist for the given `database_type`."


# Error messages related to 'save portfolio'
BAD_REQUEST_ERROR_NO_PORTFOLIO_HOLDING_ID = "Holding id is missing, specify a valid holding_id."
BAD_REQUEST_INVALID_PORTFOLIO_NAME = "Specify a valid portfolio name."
BAD_REQUEST_PORTFOLIO_NAME_ALREADY_EXISTS = "The portfolio name already exists and 'overwrite_if_exists' is False."
BAD_REQUEST_PORTFOLIO_WEIGHT_ERROR = "Holdings total weight is not 100%."
BAD_REQUEST_INVALID_PORTFOLIO_TYPE = (
    "The specified portfolio_type is not valid. Portfolio type must be either 'model_portfolios' or 'custom_benchmarks'."
)
CLIENT_ACCOUNTS_PORTFOLIO_TYPE_USED = "The 'client_accounts' portfolio_type no longer exists. Portfolios that previously had the type 'client_accounts' now have the type 'model_portfolios'"
BAD_REQUEST_INVALID_PORTFOLIO_HOLDING_DATA_FRAME = "Portfolio holdings data frame is missing required columns. Portfolio holdings data frame should have the following columns: 'Portfolio Date, 'HoldingId', 'Weight'"

# Add custom error messages here for data lookup
RESOURCE_NOT_FOUND_ERROR_DATA_SET_LOOKUP = "Requested resource does not exist for the given `universe`."
RESOURCE_NOT_FOUND_ERROR_DATA_POINT_LOOKUP = "Requested resource does not exist for the given `data_point_ids`."

# Add custom errors for environment
MALFORMED_JWT_ERROR = "Invalid JWT format. JWT should look like 'header.payload.signature' where header and payload are base64 encoded strings and signature is a hex string."

# Add custom errors for delivery
FAILED_DELIVERY_ERROR = "The delivery has failed. Please check your delivery configuration for job_id: {}. Please contact Analytics Lab support team at Morningstar for additional support."



from typing import Optional

from . import _error_messages


class UnauthorizedDataLakeAccessError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.UNAUTHORIZED_DATALAKE_REQUEST_ERROR)
        else:
            super().__init__(message)


class InvalidQueryException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class UnavailableExternally(Exception):
    def __init__(self) -> None:
        super().__init__(_error_messages.UNAVAILABLE_EXTERNALLY)


class DataNotFoundException(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(_error_messages.DATA_NOT_FOUND)


class TempTableNameNotFoundException(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(_error_messages.TEMP_TABLE_NAME_NOT_FOUND)



import warnings
from abc import ABC
from typing import Any, Callable, Dict, List, Optional, Union

import simplejson as json
from pandas import DataFrame

from .._base import _logger
from . import _decorator, _error_messages
from ._base_api import APIBackend
from ._config import _Config
from ._data_objects import TypeValue, ValidCustomDataPointTypes
from ._data_type import DatabaseCD
from ._exceptions import ResourceNotFoundError, ValueErrorException
from ._utils import _mapper_data_frame_return_dict, _mapper_data_frame_return_list
from .data_type import Blank

_config = _Config()


class CustomDatabaseAPIBackend(APIBackend):
    """
    Subclass to call the custom database API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_CUSTOM_DATABASE) from None


ResultType = Dict[str, Union[bool, str]]
ListDatapointType = List[Dict[str, Any]]
DictValues = Dict[str, TypeValue]
RelValuesWithDatapoint = Dict[str, Union[TypeValue, List[DictValues]]]


class _CustomDatabase(ABC):
    def __init__(self, database_type: DatabaseCD, custom_database_api_request: CustomDatabaseAPIBackend):
        self.database_type: DatabaseCD = database_type
        self.all_data_points: ListDatapointType = []
        self._custom_database_api_request = custom_database_api_request

    def get_data_point_from_name(self, data_point_name: str, time_series: bool = False) -> Dict[str, str]:
        data_points = self.get_all_data_points()
        data_point = next(
            (item for item in data_points if item["name"] == data_point_name),
            None,
        )

        if data_point is None or "columnType" not in data_point:
            raise ValueErrorException(f"The column '{data_point_name}' does not exist in the custom datapoints.")

        start_with_ts = data_point["columnType"].startswith("TS")

        if time_series and not start_with_ts:
            raise ValueErrorException(f"The column '{data_point_name}' is not for a time series.")

        if not time_series and start_with_ts:
            raise ValueErrorException(f"The column '{data_point_name}' is not for a single value.")

        return data_point

    def get_all_data_points(self) -> ListDatapointType:
        if not self.all_data_points:
            url = f"{_config.custom_data_points_service_url()}v1/cdp/definitions?type={self.database_type.value}"
            self.all_data_points = self._custom_database_api_request.do_get_request(url)
        return self.all_data_points

    def is_a_valid_value(self, value: Optional[TypeValue], data_point_definition: dict, investment: str) -> Optional[TypeValue]:
        if value is None:
            return value
        validator = ValidCustomDataPointTypes(data_point_definition)
        try:
            result: TypeValue = validator.is_valid(value)
            return result
        except Exception as e:
            raise ValueErrorException(f"Error in '{data_point_definition['name']}' column with '{investment}': {str(e)}")

    def convert_to_objects(self, values: DataFrame) -> Any:
        return values.astype(object).where(values.notnull(), None).where(values != "", None)

    def filter_null_values(
        self,
        values: DataFrame,
        blank: Blank,
        key_column: str,
        date_column: Optional[str] = None,
    ) -> Any:
        ops = Blank.options()
        if blank is Blank.warning and values.isnull().values.any():
            raise ValueErrorException(
                f"The dataframe contains null or blank values. \n"
                f"Please choose one of the following as a `blank` arg: {ops} \n"
            ) from None
        ignore_columns = [
            key_column,
            date_column if date_column is not None else key_column,
        ]
        if blank is Blank.ignore and values.drop(list(set(ignore_columns)), axis=1).isnull().values.all():
            raise ValueErrorException(
                f"All the dataframe values are null or blank. \n" f"Please choose one of the following as a `blank` arg: {ops} \n"
            ) from None
        return values

    def get_to_delete(self, value: TypeValue, blank: Blank) -> bool:
        if blank is Blank.ignore:
            return False
        else:
            if value is None:
                return True
        return False

    @_decorator.typechecked
    def save_historical_values(
        self,
        values: DataFrame,
        key_column: str = "SecId",
        date_column: str = "Date",
        blank: Union[Blank, str] = Blank.warning,
    ) -> ResultType:
        """Saves historical time series values for data points in the user's or firm's database in Morningstar Direct.

        Args:
            values (:obj:`DataFrame`): A DataFrame object with the time series values to be saved in the user’s or firm’s database. The DataFrame columns should include:

                * `SecId` - Column containing investment ID values. If a column named `SecId` is not present, the column name specified in the key_column argument will be used.
                * `DataPointName` - Data point name which will be updated. Multiple data points can be updated in one request.
                * `Date` - Column containing date values. If a column named `Date` is not present, the column name specified in the date_column argument will be used.

                For example::

                    ====================  ========  ===============  ===============  ===============
                    SecId                   Date    DataPointName1   DataPointName2   DataPointName3
                    ====================  ========  ===============  ===============  ===============
                    <InvestmentId>         <date>       <value>          <value>          <value>
                    <InvestmentId>         <date>       <value>          <value>          <value>
                    ====================  ========  ===============  ===============  ===============

            key_column (:obj:`str`): Name of the column containing investment ID values. If not specified, the column `SecId` will be used.

            date_column (:obj:`str`): Name of the column containing date values. If not specified, the column `Date` will be used.

            blank (:obj:`md.direct.data_type.Blank`): Argument specifying how blank values in the DataFrame should be handled. Valid enum values for this argument are:

                    * warning - Raises an error if the DataFrame contains at least one null or empty (str) value.
                    * ignore - Saves all DataFrame values except for null or empty (str) values.
                    * update - Replaces/updates all existing values with the provided values. Deletes any value that is sent as null, None or empty (str) values.

        :Returns:
            DataFrame:

        :Examples:

        ::

            import morningstar_data as md
            from pandas import DataFrame

            df = DataFrame({"SecId": ["FOUSA00C3M", "F000010NJ5"],
                            "Date": ["2022-02-10","2022-04-2"],
                            "DataPointNameText": ["2","9"],
                            "DataPointNameNumber": [12.2178, 78514]})

            md.direct.my_database.save_historical_values(values=df, key_column="SecId", blank=md.direct.data_type.Blank.warning)



        ::

            import morningstar_data as md
            from pandas import DataFrame

            df = DataFrame({"SecId": ["FOUSA00C3M", "F000010NJ5"],
                            "Date": ["2022-02-10","2022-04-2"],
                            "DataPointNameText": ["2","9"],
                            "DataPointNameNumber": [12.2178, 78514]})

            md.direct.firm_database.save_historical_values(values=df, key_column="SecId", blank=md.direct.data_type.Blank.warning)

        :Errors:
            AccessDeniedError: Raised when the user is not authenticated.

            ForbiddenError: Raised when the user does not have permission to access the requested resource.

            InternalServerError: Raised when the server encounters an unhandled error.

            NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

            ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

        """
        if not isinstance(blank, Blank):
            warnings.warn(
                "The use of string values for the 'blank' parameter is deprecated and will be removed in the next major version. Use Blank enum values instead",
                FutureWarning,
                stacklevel=2,
            )

        blank = Blank[blank]
        assert isinstance(blank, Blank)

        if key_column not in values:
            raise ValueErrorException(f"The key column '{key_column}' does not exist.")

        if date_column not in values:
            raise ValueErrorException(f"The date column '{date_column}' does not exist")

        values = self.filter_null_values(self.convert_to_objects(values), blank, key_column, date_column)

        def get_values(last_dict: Dict, column: str, value: Any, row: Dict) -> Dict:
            if column in [key_column, date_column]:
                return last_dict

            date = self.is_a_valid_value(
                row[date_column],
                {
                    "name": date_column,
                    "columnType": "Date",
                },
                row[key_column],
            )
            data_point: Dict[str, str] = self.get_data_point_from_name(column, True)

            assert isinstance(blank, Blank)
            if value is None and blank is Blank.ignore:
                return last_dict
            ts_value = self.is_a_valid_value(value, data_point, row[key_column])
            relation = f"{data_point['id']}-{row[key_column]}"

            if relation not in last_dict:
                last_dict[relation] = []

            to_delete = self.get_to_delete(ts_value, blank)
            last_dict[relation].append({"date": date, "value": ts_value, "toDelete": to_delete})
            return last_dict

        relations: DictValues = _mapper_data_frame_return_dict(values, get_values)
        data: List[RelValuesWithDatapoint] = [
            {
                "dataPointId": relation.split("-")[0],
                "investmentId": relation.split("-")[1],
                "timeSeriesValue": points,
            }
            for relation, points in relations.items()
        ]
        return self._request_save_values(data)

    @_decorator.typechecked
    def save_values(
        self,
        values: DataFrame,
        key_column: str = "SecId",
        blank: Union[Blank, str] = Blank.warning,
    ) -> ResultType:
        """Saves values for data points in the user's or firm's database.

        Args:
            values (:obj:`DataFrame`): A DataFrame object with the values to be saved in the user's or firm's database. The DataFrame columns should include:

                * `SecId` - Column containing investment ID values. If a column named `SecId` is not present, the column name specified in the key_column argument will be used.
                * `DataPointName` - Data point name which will be updated. Multiple data points can be updated in one request.

                For example::

                    ====================  ===============  ===============  ===============
                    SecId                 DataPointName1   DataPointName2   DataPointName3
                    ====================  ===============  ===============  ===============
                    <InvestmentId>            <value>          <value>          <value>
                    <InvestmentId>            <value>          <value>          <value>
                    ====================  ===============  ===============  ===============

            key_column (:obj:`str`): Name of the column containing investment ID values. If not specified, the column `SecId` will be used.

            blank (:obj:`md.direct.data_type.Blank`): Argument specifying how blank values in the DataFrame should be handled. Valid enum values for this argument are:

                * warning - Raises an error if the DataFrame contains at least one null or empty (str) value.
                * ignore - Saves all DataFrame values except for null or empty (str) values.
                * update - Replaces/updates all existing values with the provided values. Deletes any value that is sent as null, None or empty (str) values.
        :Returns:
            DataFrame:

        :Examples:

        ::

            import morningstar_data as md
            from pandas import DataFrame
            df = DataFrame({"SecId": ["FOUSA00C3M", "F000010NJ5"],
                               "DataPointNameText": ["2","9"],
                               "DataPointNameDate": ["2022-02-10","2022-04-2"],
                               "DataPointNameNumber": [12.2178, 78514]})
            md.direct.my_database.save_values(values=df, key_column="SecId", blank=md.direct.data_type.Blank.warning)

        ::

            import morningstar_data as md
            from pandas import DataFrame
            df = DataFrame({"SecId": ["FOUSA00C3M", "F000010NJ5"],
                               "DataPointNameText": ["2","9"],
                               "DataPointNameDate": ["2021-05-31","2021-02-01"],
                               "DataPointNameNumber": [12.2178, 78514]})
            md.direct.firm_database.save_values(values=df, key_column="SecId", blank=md.direct.data_type.Blank.warning)

        :Errors:
            AccessDeniedError: Raised when the user is not authenticated.

            ForbiddenError: Raised when the user does not have permission to access the requested resource.

            InternalServerError: Raised when the server encounters an unhandled error.

            NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

            ResourceNotFoundError: Raised when the requested resource does not exist in Direct.
        """
        if not isinstance(blank, Blank):
            warnings.warn(
                "The use of string values for the 'blank' parameter is deprecated and will be removed in the next major version. Use Blank enum values instead",
                FutureWarning,
                stacklevel=2,
            )

        blank = Blank[blank]
        assert isinstance(blank, Blank)

        if key_column not in values:
            raise ValueErrorException(f"The key column '{key_column}' does not exist.")

        values = self.filter_null_values(self.convert_to_objects(values), blank, key_column)

        def get_values(column: str, value: Any, row: dict) -> Optional[RelValuesWithDatapoint]:
            if column is key_column:
                return None

            assert isinstance(blank, Blank)
            if value is None and blank is Blank.ignore:
                return None
            data_point = self.get_data_point_from_name(column)
            single_value = self.is_a_valid_value(value, data_point, row[key_column])
            return {
                "dataPointId": data_point["id"],
                "investmentId": row[key_column],
                "singleValue": single_value,
                "toDelete": self.get_to_delete(value, blank),
            }

        data: List[RelValuesWithDatapoint] = _mapper_data_frame_return_list(values, get_values)

        return self._request_save_values(data)

    def _request_save_values(self, data: List[RelValuesWithDatapoint]) -> ResultType:
        url = f"{_config.custom_data_points_service_url()}v1/cdp/investments/values?type={self.database_type.value}"
        result: ResultType = self._custom_database_api_request.do_put_request(url, json.dumps(data, ignore_nan=True))
        return result


my_database = _CustomDatabase(DatabaseCD.user, CustomDatabaseAPIBackend())
firm_database = _CustomDatabase(DatabaseCD.firm, CustomDatabaseAPIBackend())



"""
Here lie methods to interact with the Morningstar Data API. These methods are not intended to be used directly by
users of the library. They will be used by the public methods in the morningstar_data package.
"""

import dataclasses
import io
import os
import time
import uuid
from typing import Dict

import pandas as pd
import requests
import urllib3

from .._utils import join_url
from .._version import __version__
from ..direct._config import _Config
from ._exceptions import (
    MdApiRequestException,
    MdApiTaskException,
    MdApiTaskTimeoutException,
    MdBaseException,
    exception_by_name,
)
from ._types import MdapiTask, RequestObject, TaskResult

_config = _Config()

# Timeouts will generally be managed in the API, but just in case, let's have a fallback super-long timeout here
POLL_TIMEOUT = int(os.environ.get("MD_TASK_TIMEOUT_SECONDS", 10_000))

MD_API_V1_BASE = join_url(os.environ.get("MD_API", "https://www.us-api-proxy.morningstar.com/md-api"), "/v1")
MD_API_REQUEST_TIMEOUT = 5  # seconds
S3_CONNECTION_TIMEOUT = 5  # seconds
S3_READ_TIMEOUT = 60  # seconds


def _get_headers() -> Dict[str, str]:
    request_id = str(uuid.uuid4())
    auth_token = _config.get_uim_token()

    headers = {
        "authorization": f"Bearer {auth_token}",
        "md-correlation-id": request_id,
        "md-request-id": request_id,
        "md-package-version": __version__,
        "md-direct-source-app": os.getenv("DO_API_REQUEST_ORIGIN", "morningstar-data"),
    }
    if os.getenv("FEED_RUN_ID"):
        headers["md-feed-run-id"] = os.getenv("FEED_RUN_ID")

    if os.getenv("REQUEST_ID"):  # When running on al-api pods, this env var will be present
        headers["md-feed-run-attempt-id"] = os.getenv("REQUEST_ID")
        headers["md-correlation-id"] = os.getenv("REQUEST_ID")

    if os.getenv("FEED_ID"):
        headers["md-feed-id"] = os.getenv("FEED_ID")

    return headers


def _read_feather_from_s3(url: str) -> pd.DataFrame:
    if url == "":
        return pd.DataFrame()
    download_response = requests.get(url, timeout=(S3_CONNECTION_TIMEOUT, S3_READ_TIMEOUT))
    download_response.raise_for_status()
    feather_content = io.BytesIO(download_response.content)
    return pd.read_feather(feather_content)


def _wait_for_completed_task_response(task: MdapiTask) -> MdapiTask:
    task_status_request_session = requests.Session()
    task_status_request_session.mount(
        "https://",
        requests.adapters.HTTPAdapter(
            max_retries=urllib3.util.Retry(total=3, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
        ),
    )

    task_start_time = int(time.time())
    while not task.is_complete():
        if time.time() - task_start_time > POLL_TIMEOUT:
            raise MdApiTaskTimeoutException(task_id=task.id, timeout_time=POLL_TIMEOUT)
        time.sleep(task.next_poll_delay_milliseconds / 1000.0)
        task = _check_task_status(task.poll_url, task_start_time, task_status_request_session)
    return task


def _check_task_status(polling_url: str, task_start_time: float, task_status_request_session: requests.Session) -> MdapiTask:
    request_headers = _get_headers()
    try:
        response = task_status_request_session.get(
            polling_url,
            headers=request_headers,
            timeout=MD_API_REQUEST_TIMEOUT,
            params={"task_start_time": task_start_time},
        )
        response.raise_for_status()
    except requests.RequestException as e:
        raise MdApiRequestException(request_id=request_headers["md-request-id"], url=polling_url, detail=str(e)) from e

    return MdapiTask(**response.json())


def _create_task(url: str, request_object: RequestObject) -> MdapiTask:
    request_headers = _get_headers()
    try:
        task_response = requests.post(
            url, json=dataclasses.asdict(request_object), headers=request_headers, timeout=MD_API_REQUEST_TIMEOUT
        )
        task_response.raise_for_status()
    except requests.HTTPError:
        request_id = request_headers["md-request-id"]
        try:
            response_json = task_response.json()
        except requests.exceptions.JSONDecodeError:
            message = f"Unexpected API error format. Response body: '{task_response.text}'"
            raise MdApiRequestException(request_id=request_id, url=url, detail=message) from None
        raise exception_by_name(
            response_json.get("error_class", ""), request_id=request_id, message=response_json.get("error_message", "")
        ) from None
    except requests.RequestException as e:
        raise MdApiRequestException(request_id=request_headers["md-request-id"], url=url, detail=str(e)) from e

    return MdapiTask(**task_response.json())


def call_remote_function(method: str, request_object: RequestObject) -> pd.DataFrame:
    url = join_url(MD_API_V1_BASE, method)

    task = _create_task(url, request_object)
    try:
        response = _wait_for_completed_task_response(task)
        # The API sets the dataframe_file_url if a dataframe is empty
        if response.is_successful():
            # response.result should always be present, but check for mypy's sake
            task_result = response.result or TaskResult(dataframe_file_url="")
            df = _read_feather_from_s3(task_result.dataframe_file_url)
            for alias in task_result.columns_with_list_values:
                # Maintaining some legacy behavior here. If the column has only one value, we'll convert it to a singular value.
                # We may want to consider changing this as a breaking change in the future.
                df[alias] = df[alias].apply(lambda x: x[0] if len(x) == 1 else list(x))
            return df

        if response.failed():
            raise exception_by_name(str(response.error_class), task_id=task.id, message=str(response.error_message)) from None

    # Let's catch any unhandled exceptions and attach the task id so it can be shared for troubleshooting help.
    # We'll ignore our custom exceptions since they already have all the detail we would want.
    except MdBaseException:
        raise
    except Exception as e:  # pylint: disable=broad-exception-caught
        raise MdApiTaskException(task_id=str(task.id), detail=repr(e)) from e



import uuid
from typing import Any, Dict, List, Optional, Union

import requests
import simplejson as json
from pandas import DataFrame

from ..._base import _logger
from .. import _decorator, investment
from .._base_api import APIBackend
from .._config import _Config
from .._error_messages import (
    BAD_REQUEST_ERROR_INVALID_INVESTMENT_LIST_ID,
    RESOURCE_NOT_FOUND_ERROR_INVESTMENT_List,
)
from .._exceptions import (
    ApiResponseException,
    BadRequestException,
    NetworkExceptionError,
    ResourceNotFoundError,
    ValueErrorException,
)
from .._utils import _reduce_list_data

_config = _Config()


class InvestmentListAPIBackend(APIBackend):
    """
    Subclass to call the Holding API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(RESOURCE_NOT_FOUND_ERROR_INVESTMENT_List) from None


_investment_list_api_request = InvestmentListAPIBackend()


@_decorator.not_null
@_decorator.typechecked
def get_investment_list(list_id: str) -> DataFrame:
    """Returns all investments that belong to a saved investment list in Morningstar Direct.

    Args:
        list_id (:obj:`str`): The unique identifier of a saved investment list in Morningstar
            Direct, e.g., "EBE416A3-03E0-4215-9B83-8D098D2A9C0D". Use the `get_investment_lists <#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.

    :Returns:
        DataFrame: A DataFrame object with investments that belong to the specified list. DataFrame columns include:

        * secid
        * masterportfolioid
        * tradingsymbol
        * name
        * securitytype
        * exchangeid
        * category

    :Examples:
        Get investment list by ID.

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_investment_list(list_id="385349FE-01D6-4064-B297-64EAA28BD4E9") # Replace with a valid List ID
        df

    :Output:
        ==========  =================  =============  =============================  ============  ==========  ========
        secid       masterportfolioid  tradingsymbol  name                           securitytype  exchangeid  category
        ==========  =================  =============  =============================  ============  ==========  ========
        XIUSA000KQ  24729              NaN            Russell 2000 Growth TR USD XI  NaN           Small       Growth
        ...
        ==========  =================  =============  =============================  ============  ==========  ========

    """
    try:
        # sourcetype should be "Desktop" to retrieve secids for list items
        # The alternative is "Cloud", which retrieves performanceids.
        # We need secids for joining with Datalake tables
        _validate_investment_list_id(list_id)
        _logger.debug(f"Getting Direct List: {list_id}")
        url = f"{_config.object_service_url()}objectapi/v1/lists/{list_id}?sourcetype=Desktop"
        response_json: Dict[str, Any] = _investment_list_api_request.do_get_request(url)

        # The Lists API will respond with a 200 if the list is not found.
        # Instead of relying on HTTP status code, let's check their JSON response for any errors
        if response_json["_meta"]["response_status"] != "200100":
            raise ApiResponseException(response_json["_meta"]["hint"])

        list_json: Dict[str, Any] = response_json["lists"][0]

        ids = list(map(lambda x: f"{x['id']}", list_json["idList"]))
        ids_types = list(map(lambda x: f"{x['3xType']}", list_json["idList"]))

        if not ids:
            list_df = DataFrame(
                {
                    "secid": [],
                    "masterportfolioid": [],
                    "tradingsymbol": [],
                    "name": [],
                    "securitytype": [],
                    "exchangeid": [],
                    "category": [],
                }
            )
        else:
            list_df = investment.get_investment_data(
                investments=ids,
                data_points=[
                    {"datapointId": "OS00I", "isTsdp": False},
                    {"datapointId": "DC09A", "isTsdp": False},
                    {"datapointId": "OS385", "isTsdp": False},
                    {"datapointId": "OS01W", "isTsdp": False},
                    {"datapointId": "OS010", "isTsdp": False},
                    {"datapointId": "OS01Z", "isTsdp": False},
                    {"datapointId": "OS245", "isTsdp": False},
                ],
            )
            list_df = list_df.drop(["Id"], axis=1)
            list_df.columns = [
                "secid",
                "masterportfolioid",
                "tradingsymbol",
                "name",
                "securitytype",
                "exchangeid",
                "category",
            ]
            list_df["secid"] = ids
            list_df["securitytype"] = ids_types
        # We are getting float values for masterportfolioid from the api and the fact that this statement is trying
        # to convert it to integer and convert Na/NaN values to python None object.
        list_df["masterportfolioid"] = (
            list_df["masterportfolioid"].fillna(-1).astype(int).astype(str).replace({"-1": None, "None": None})
        )

        if list_df.empty:
            _logger.debug(f"Returning empty DataFrame.")
        _logger.debug(f"Returning DataFrame for List: {list_id}")
        return list_df

    except Exception as e:
        _logger.error(repr(e))
        error_message = str(e)
    if error_message:
        raise ValueErrorException(error_message)


@_decorator.typechecked
def get_investment_lists() -> DataFrame:
    """Returns all investment lists saved or shared to a user in Morningstar Direct. Also includes Morningstar
    pre-defined investment lists.

    :Returns:
        DataFrame: A DataFrame object containing investment list details. DataFrame columns include:

        * id
        * name

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_investment_lists()
        df

    :Output:
        ====================================  ==============================
        id                                    name
        ====================================  ==============================
        EBE416A3-03E0-4215-9B83-8D098D2A9C0D  Morningstar Open Index Project
        858BD493-68B3-4D44-9DF0-333D3CC88A1C  Morningstar Prospects
        ...
        ====================================  ==============================

    """
    try:
        _logger.debug(f"Getting All Direct Lists")
        url = f"{_config.object_service_url()}objectapi/v1/lists"
        response_json: Dict[str, Any] = _investment_list_api_request.do_get_request(url)

        # The Lists API will respond with a 200 if the list is not found.
        # Instead of relying on HTTP status code, let's check their JSON response for any errors
        if response_json["_meta"]["response_status"] != "200100":
            raise ApiResponseException(response_json["_meta"]["hint"])

        props_dict: Dict[str, str] = {"id": "id", "name": "name"}
        all_investment_list: List[Dict[str, Any]] = _reduce_list_data(response_json["lists"], props_dict)
        return DataFrame(all_investment_list)
    except Exception as e:
        _logger.error(repr(e))
        error_message = str(e)
    if error_message:
        raise ValueErrorException(error_message)


@_decorator.typechecked
def save_investment_list(list_name: str, investment_ids: List[str], overwrite_if_exists: bool) -> DataFrame:
    """Saves or updates an investment list.

    Args:
        list_name (:obj:`str`): Name of the list
        investment_ids (:obj:`list`): List of investment IDs (SecId). For example: ["F00000YVYF","FOUSA00CFV"]
        overwrite_if_exists(:obj:`bool`): If True the list will be overwritten

    :Returns:
        DataFrame: A DataFrame object with the new investment list data.

    :Examples:

    Example 1: Save a new investment list.

    Output DataFrame columns include:

    * Status
    * Name
    * List Id
    * Investments
    * Created Date

    ::

        import morningstar_data as md

        df = md.direct.user_items.save_investment_list(
            list_name="new_investments",
            investment_ids=["F00000YVYF", "FOUSA00CFV"],
            overwrite_if_exists=False
        )
        df

    :Output:
        =========================  ===============  ====================================  ========================  ===================
        Status                     Name             List Id                               Investments               Created Date
        =========================  ===============  ====================================  ========================  ===================
        List successfully created  new_investments  295F7E59-15AC-4424-958E-3BD8B0A733EE  [F00000YVYF, FOUSA00CFV]  2022-03-29T14:43:00
        ...
        =========================  ===============  ====================================  ========================  ===================

        Example 2: Update an existing list.

        Output DataFrame columns include:

        * Status
        * Name
        * List Id
        * Modified Date

        ::

            import morningstar_data as md

            df = md.direct.user_items.save_investment_list(
                list_name="new_investments",
                investment_ids=["F00000YVYF", "FOUSA00CFV"],
                overwrite_if_exists=True
            )
            df


        =========================  ===============  ====================================  ===================
        Status                     Name             List Id                               Modified Date
        =========================  ===============  ====================================  ===================
        List successfully updated  new_investments  295F7E59-15AC-4424-958E-3BD8B0A733EE  2022-03-29T14:45:00
        ...
        =========================  ===============  ====================================  ===================

    """

    try:
        if list_name is None or list_name.strip() == "":
            raise BadRequestException("Please specify a list name to proceed.") from None
        if not investment_ids:
            raise BadRequestException("Please specify investment ids to proceed.") from None

        get_item = get_investment_lists()

        data: Dict[str, Any]
        if (list_name in get_item.values) and overwrite_if_exists == False:
            raise BadRequestException("The list name already exists.") from None
        if (list_name in get_item.values) and overwrite_if_exists:
            list_id = get_item.loc[get_item["name"] == list_name, "id"].iloc[0]
            url = f"{_config.object_service_url()}objectapi/v1/lists/{list_id}?sourcetype=Desktop"
            data = {"name": list_name, "listType": "ID_LIST", "idList": investment_ids}
            result = _investment_list_api_request.do_put_request(url, json.dumps(data, ignore_nan=True))
            if result["_meta"]["response_status"] != "200102":
                raise ApiResponseException(result["_meta"]["hint"]) from None
            data = {
                "Status": [result["_meta"]["hint"]],
                "Name": [result["lists"][0]["name"]],
                "List Id": [result["lists"][0]["id"]],
                "Modified Date": [result["lists"][0]["lastModifiedOn"]],
            }
            return DataFrame(data)
        else:
            url = f"{_config.object_service_url()}objectapi/v1/lists?sourcetype=Desktop"
            data = {"name": list_name, "listType": "ID_LIST", "idList": investment_ids}
            result = _investment_list_api_request.do_post_request(url, json.dumps(data, ignore_nan=True))
            ids = list(map(lambda x: f"{x['id']}", result["lists"][0]["idList"]))
            data = {
                "Status": [result["_meta"]["hint"]],
                "Name": [result["lists"][0]["name"]],
                "List Id": [result["lists"][0]["id"]],
                "Investments": [ids],
                "Created Date": [result["lists"][0]["createdOn"]],
            }
            return DataFrame(data)
    except Exception as e:
        _logger.error(repr(e))
        error_message = str(e)
        if error_message:
            raise ValueErrorException(error_message)


def _validate_investment_list_id(list_id: str) -> None:
    try:
        uuid.UUID(list_id)
    except ValueError:
        _logger.debug(f"Investment list id {list_id} is not valid.")
        raise BadRequestException(BAD_REQUEST_ERROR_INVALID_INVESTMENT_LIST_ID) from None



from typing import Any, Dict, Optional, Union

import requests

from ..._base import _logger
from .._base_api import APIBackend
from .._config import _Config
from .._core_api import make_request
from .._error_messages import RESOURCE_NOT_FOUND_ERROR_HOLDING
from .._exceptions import (
    BadRequestException,
    ForbiddenError,
    NetworkExceptionError,
    ResourceNotFoundError,
)

_config = _Config()


class HoldingAPIBackend(APIBackend):
    """
    Subclass to call the Holding API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(RESOURCE_NOT_FOUND_ERROR_HOLDING) from None


class FoFAPIBackend(APIBackend):
    """
    Subclass to call the FoF API and handle any HTTP errors that occur.
    Temporarily implements do_get_request and do_put_request becuase the response of FOF api is not JSON
    """

    def __init__(self) -> None:
        super().__init__()

    def do_get_request(self, url: str) -> Any:
        """
        Makes a GET request
        """
        return self._make_request(url, "GET")

    def do_post_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a POST request
        """
        return self._make_request(url, "POST", data)

    def _make_request(self, url: str, method: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Request and handle OK (status_code = 200) and network error responses.
        """
        _logger.info(f"Requesting: {method} {url}")
        try:
            res = make_request(method, url, headers=self.get_headers(), verify=True, data=data)
            res.raise_for_status()
            # We have asked FOF team to return JSON object in the API, so we don't have to implement this class completely like this.
            # https://mswiki.morningstar.com/display/DWH/AL+FOF+Lookthrough+Holding?focusedCommentId=645255358#comment-645255358
            return res.json()
        except requests.ConnectionError:
            raise NetworkExceptionError from None
        except requests.HTTPError:
            self._handle_custom_http_errors(res)
            self._handle_default_http_errors(res)

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.text
        _logger.error(response_message)
        status_code = res.status_code

        if status_code in [404, 500]:
            raise NetworkExceptionError from None
        elif status_code == 401:
            raise ForbiddenError from None
        else:
            raise BadRequestException(response_message)

    def _handle_default_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with standard messages
        """
        response_message = res.text
        _logger.error(response_message)
        raise


class AMSAPIBackend(APIBackend):
    """
    Subclass to call a Signed url and handle any HTTP errors that occur.
    Temporarily implements do_get_request becuase the request does not need headers and the response need not be JSON
    """

    def __init__(self) -> None:
        super().__init__()

    def do_get_request(self, url: str) -> Any:
        """
        Makes a GET request
        """
        return self._make_request(url, "GET")

    def _make_request(self, url: str, method: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Request and handle OK (status_code = 200) and network error responses.
        """
        _logger.info(f"Requesting: {method} {url}")
        try:
            res = make_request(method, url, headers=self.get_headers(), verify=True, data=data)
            res.raise_for_status()
            return res.json()
        except requests.ConnectionError:
            raise NetworkExceptionError from None
        except requests.HTTPError:
            self._handle_custom_http_errors(res)
            self._handle_default_http_errors(res)

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(RESOURCE_NOT_FOUND_ERROR_HOLDING) from None
